/* ************************************************************* *
CHIPNSFX interactive three-channel musical tracker and data editor
written by Cesar Nicolas-Gonzalez / CNGSOFT since 2017-02-09 14:20
* ************************************************************* */

#define MY_FILE "CHIPNSFX"
#define MY_DATE "20201031"//"2555"
#define MY_NAME "CHIPNSFX tracker"
#define MY_SELF "Cesar Nicolas-Gonzalez / CNGSOFT"

#define GPL_3_INFO "This program comes with ABSOLUTELY NO WARRANTY; for more details" "\n" \
	"please read the GNU General Public License. This is free software" "\n" \
	"and you are welcome to redistribute it under certain conditions."

/* This notice applies to the sources of CHIPNSFX and its binaries.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

Contact information: <mailto:cngsoft@gmail.com> */

#include <stdio.h>
#include <ctype.h> // toupper...
#include <stdlib.h> // malloc...
#include <string.h> // strcpy...
#ifdef _WIN32
	#define strcasecmp _stricmp
#endif
#ifndef DEBUG
	#define WIN32_LEAN_AND_MEAN 1
	#include <windows.h> // console input+output operations
	#include <mmsystem.h> // audio output; requires -lwinmm
#endif
#define INLINE // `inline`: ignored by TCC and GCC4, harmful in GCC5!

/*--------*/

#define MAXSTRLEN 1024 // a QUANTUM (44100/100=441) or a pattern ("#01 A-4:01 ...":7*(1+128)=903) must fit inside
#define CHIPNSFX_SIZE 0x70

static char notes1[]="CCDDEFFGGAAB",notes2[]="-#-#--#-#-#-"; // C- C# D- D# E- F- F# G- G# A- A# B-
char pianos[]="zsxdcvgbhnjm,q2w3er5t6y7ui9o0pl1a."; // standard QWERTY keyboard: 12+12+6+SFX+REST+BRAKE+ERASE
static char pianoz[][sizeof(pianos)]={
	"ysxdcvgbhnjm,q2w3er5t6z7ui9o0pl1a.", // QWERTZ
	"wsxdcvgbhnj,;a\202z\042er(t-y\212ui\207o\205pl&q:" // AZERTY: beware of OEM rather than ANSI or UTF8!
	};
static char chipnsfxid[]=MY_FILE " format\n",chipnsfxlf[]="===\n",vubars[2][16]={"_.......oooooOO@","0123456789ABCDEF"},crlf[]="\n";

// song file header
int hertz,divisor,transpose,loopto,orders,instrs,loopto_,orders_,instrs_;
// song file body
unsigned char t[MAXSTRLEN],
	title0[MAXSTRLEN],
	title1[MAXSTRLEN],title2[MAXSTRLEN],
	filepath[MAXSTRLEN],
	order[3][256][2],
	order_[3][256][2], // 3 channels * 256 orders * index & transposition
	scrln[256],
	scrln_[256], // 256 score pattern lengths ranging from 0 (void) to 128 (full)
	score[256][128][2],
	score_[256][128][2], // 256 score patterns * 128 rows * note+instrument
	instr[256][25],
	instr_[256][25], // 256 instruments * 8 bytes+16 ASCIZ
	clipboard[3][256][2];
FILE *f; // I/O operations never require more than one file

int flag_b=0,flag_v=0,flag_c=80,flag_r=0,flag_m=0,flag_pp=0,flag_a=0,clipboards=0,clipboardtype=0,clipboarditems=0,flag_q=0,flag_bb[3]={0,1,2},flag_t=0,flag_zz=0,flag_n=0,flag_nn=0,flag_xx=0;

/*--------*/

INLINE int intmin(int x,int y) { return x<y?x:y; }
INLINE int intmax(int x,int y) { return x>y?x:y; }
#define INTMINMAX(x,y,a,b) (x=intmin(a,b),y=intmax(a,b))
INLINE int itoradix(int x) { return x<10?x+48:((x-9)+64); } // flawed on purpose:
INLINE int radixtoi(int x) { return x<65?x-48:((x+9)&15); } // allowing ":" as 10
INLINE void keepwithin(int *i,int n,int x) { if (*i<n) *i=n; else if (*i>x) *i=x; }

char *strcln(char *s) // remove trailing whitespaces
{
	if (!s)
		return NULL;
	char *r=s;
	while (*r)
		++r;
	while (r>s&&(*--r<33))
		*r=0;
	return s;
}
char *strpad(char *s,int l) // pad string with up to l spaces
{
	char *r=s;
	while (l&&*r)
		++r,--l;
	while (l--)
		*r++=' ';
	*r=0;
	return s;
}

char musical_[]="::::::";
char *musical(int n,int i) // generate "A-4:01"
{
	if (n>0&&n<=(CHIPNSFX_SIZE-3))
	{
		--n;
		musical_[0]=notes1[n%12];
		musical_[1]=notes2[n%12];
		musical_[2]=itoradix(n/12);
	}
	else
	{
		switch (n)
		{
			case CHIPNSFX_SIZE-2:
				n='='; break;
			//case CHIPNSFX_SIZE-1:
				//n='?'; break;
			case CHIPNSFX_SIZE-0:
				n='^'; break;
			default:
				n='.'; break;
		}
		musical_[0]=musical_[1]=musical_[2]=n;
		i=0;
	}
	if (i)
		musical_[4]=itoradix(i/16),musical_[5]=itoradix(i%16);
	else
		musical_[4]=musical_[5]='.';
	return musical_;
}
INLINE int addnote(int n,int d) // transpose+normalize
{
	if (n<1)
		return 0; // empty
	if (n>=(CHIPNSFX_SIZE-0))
		return (CHIPNSFX_SIZE-0); // rest
	if (n>(CHIPNSFX_SIZE-4))
		return n; // SFX
	if ((n+=d)<1)
		return 1; // C-0
	if (n>(CHIPNSFX_SIZE-4))
		return (CHIPNSFX_SIZE-4); // B-8
	return n;
}

int flag_q_last,flag_q_next;
#define flag_q_init(i) flag_q_last=(i)+1
void flag_q_loop(int i)
{
	if (!flag_q)
		if (!(7&++flag_q_next))
			fprintf(stderr,"%02i%%\x0D",(100*i)/flag_q_last);
}
void flag_q_exit(char *s)
{
	if (!flag_q)
		fprintf(stderr,"%s\n",s);
}

/*--------*/

#define MEMVOID(v) memset((v),0,sizeof((v)))
#define MEMFULL(v) memset((v),-1,sizeof((v)))
#define MEMLOAD(v,w) memcpy((v),(w),sizeof((w)))

void newsong(void) // destroys all song data and starts anew
{
	hertz=50; divisor=6; transpose=0; loopto=-1;
	strcpy(title1,"Song title"); strcpy(title2,"Description");
	MEMVOID(order); MEMVOID(score); MEMVOID(scrln); MEMVOID(instr);
	order[1][0][0]=scrln[0]=scrln[1]=scrln[2]=orders=1;
	order[2][0][0]=instrs=2;
	instr[1][0]=instr[1][1]=-1; strcpy(&instr[1][8],"Piano");
}

/*--------*/

int crlftype=0;
char *fgets0(char *t,int x) // fgets() + cleanup
{
	if (!fgets(t,x,f))
		return NULL;
	char *s=t;
	while (*s)
		++s;
	while (s>=t&&*s<33)
		--s;
	s[1]='\n'; s[2]=0;
	return t;
}

int loadsong(char *s) // destroys the old song and loads a new one; 0 is ERROR
{
	char *r;
	int i,j; f=stdin;
	if (strcmp("-",s))
		if (!*s||!(f=fopen(s,"r")))
			return 0;
	fgets0(t,MAXSTRLEN);
	if (strcmp(t,chipnsfxid))
	{
		if (f!=stdin)
			fclose(f);
		return 0;
	}
	// header: text+vars
	newsong();
	if (!strcln(fgets0(title1,MAXSTRLEN)))
		return 0;
	if (!strcln(fgets0(title2,MAXSTRLEN)))
		return 0;
	fgets0(t,MAXSTRLEN); // skip the separator gap
	// "hertz/divisor +trasposition"
	fgets0(t,MAXSTRLEN);
	hertz=strtol(t,&r,10);
	divisor=strtol(++r,&r,10);
	transpose=strtol(r,NULL,10)+flag_nn;
	if (flag_v)
		printf("%s --- %s\n%i/%i Hz %+i Trans\nOrders:",title1,title2,hertz,divisor,transpose);
	// orders: "#00+00 #01+00 #02+12"
	i=0;
	for (;;)
	{
		if (!strcln(fgets0(t,MAXSTRLEN)))
			return 0;
		if (*(r=t)!='#')
			break;
		for (j=0;j<3;++j)
		{
			order[j][i][0]=strtol(++r,&r,16); // skip '#'
			order[j][i][1]=strtol(r,&r,10);
			++r; // skip space
		}
		if (*r)//=='*'
			loopto=i;
		if (flag_v)
			printf(" %02X",i);
		++i;
	}
	if ((orders=i)>255)
		return 1; // overrun!
	// scores: "#00: A-4:01 ...:..", etc.
	if (flag_v)
		printf("\nScores:");
	for (;;)
	{
		if (!strcln(fgets0(t,MAXSTRLEN)))
			return 0;
		if (*(r=t)!='#')
			break;
		i=strtol(++r,&r,16)&255; // overrun?
		j=0;
		for (;;)
		{
			while (*r&&(*r<33))
				++r;
			if (!*r)
				break;
			int n;
			if (n=(int)strchr(notes1,*r))
			{
				if (*++r=='+')
				{
					++r;
					n=score[i][j-1][0]+12*(radixtoi(toupper(*r++)-4));
				}
				else
				{
					if (*r++=='#')
						++n;
					n+=12*(radixtoi(toupper(*r++)))-(int)notes1+1;
				}
				if (n<1||n>(CHIPNSFX_SIZE-3))
					n=CHIPNSFX_SIZE-3; // old C-B -> new C-9
				score[i][j][1]=strtol(++r,0,16);
			}
			else
			{
				n=(*r=='^'?(CHIPNSFX_SIZE-0):(*r=='='?(CHIPNSFX_SIZE-2):0));
				score[i][j][1]=0;
			}
			score[i][j][0]=n;
			++j;
			while (*r&&!(*r<33))
				++r;
		}
		scrln[i]=j;
		if (flag_v)
			printf(" %02X",i);
	}
	// instruments: ":02 FFFE 0000 141 Ding"
	if (flag_v)
		printf("\nInstrs:");
	instrs=0;
	for (;;)
	{
		if (!strcln(fgets0(t,MAXSTRLEN)))
			return 0;
		if (*(r=t)!=':')
			break;
		if ((i=strtol(++r,&r,16))>0&&i<255)
		{
			j=strtol(++r,&r,16);
			instr[i][0]=j>>8; instr[i][1]=j;
			j=strtol(++r,&r,16);
			instr[i][2]=j>>8; instr[i][3]=j;
			j=strtol(++r,&r,16);
			instr[i][4]=j>>8; instr[i][5]=j;
			if (*r)
				++r;
			strcpy(&instr[i][8],r);
			if (instrs<i)
				instrs=i;
			if (flag_v)
				printf(" %02X",i);
		}
	}
	if (++instrs>255)
		return 1; // overrun!
	for (i=0;i<256;++i)
		for (j=0;j<128;++j)
			if (score[i][j][1]>instrs)
				score[i][j][1]=score[i][j][0]=0; // unknown instrument = nothing
	if (flag_v)
		printf(crlf);
	if (f!=stdin)
		fclose(f);
	strcpy(filepath,s);
	return 1;
}

/*--------*/

int savesong(char *s) // saves the current song onto a file; 0 is ERROR
{
	int i,j,k;
	if (!s||!(f=fopen(s,"w")))
		return 0;
	fprintf(f,"%s%s\n%s\n%s%i/%i %+03i\n",chipnsfxid,title1,title2,chipnsfxlf,hertz,divisor,transpose-flag_nn);
	for (i=0;i<orders;++i)
		fprintf(f,"#%02X%+03i #%02X%+03i #%02X%+03i%s\n",
			order[0][i][0],(signed char)order[0][i][1],
			order[1][i][0],(signed char)order[1][i][1],
			order[2][i][0],(signed char)order[2][i][1],
			i==loopto?" *":"");
	fputs(chipnsfxlf,f);
	for (i=0;i<256;++i)
	{
		k=0;
		for (j=0;j<orders;++j)
			k|=(order[0][j][0]==i)||(order[1][j][0]==i)||(order[2][j][0]==i);
		if (k) // skip unused scores!
		{
			k=scrln[i];
			fprintf(f,"#%02X",i);
			for (j=0;j<k;++j)
				fprintf(f," %s",musical(score[i][j][0],score[i][j][1]));
			fputs(crlf,f);
		}
	}
	fputs(chipnsfxlf,f);
	for (i=1;i<instrs;++i)
	{
		sprintf(t,":%02X %02X%02X %02X%02X %01X%02X %s",i,
			instr[i][0],instr[i][1],
			instr[i][2],instr[i][3],
			instr[i][4],instr[i][5],
			&instr[i][8]);
		j=strlen(t);
		while (t[--j]==' ')
			{}
		t[++j]='\n';
		t[++j]=0;
		fputs(t,f);
	}
	fputs(chipnsfxlf,f);
	fclose(f);
	return 1;
}

/*--------*/

int save_hits[256],save_chan[256],save_tiny[256],savesizes,save_full[3];
int play_pose,past_pose,play_size,past_size,save_size[256],play_flag[6],past_flag[6],save_flag[256][6]; // <0 unused, 0..255 single, 256 multi.
char save_pack[256],save_packscore[256][128],save_packflags[256][128]; int save_packorder[3][256]; // +n = length, -n = -offset
int savechar;
#define savedefb(q) if (q) savechar=0
char *savegaps(void)
{
	if (savechar>=flag_c)
		return savechar=0,"\n  db "; // two spaces
	else
		return savechar>0?",":" db "; // one space
}
#define savecrlf(q) if (q) fputs(crlf,f)

int saveplayback(int n,int i) // note,instrument
{
	if ((n>0&&n<CHIPNSFX_SIZE-2)) // note or SFX?
	{
		if (i) // instrument?
		{
			if (play_flag[0]=instr[i][0]) // mute? ignore volume envelope
				play_flag[1]=instr[i][1];
			if (play_flag[2]=instr[i][2]) // clear? ignore noise envelope
				play_flag[3]=instr[i][3];
			if (n<(CHIPNSFX_SIZE-3)) // SFX: ignore
				play_flag[4]=(play_flag[5]=instr[i][5])?instr[i][4]:0; // force type 0 on empty SFX
		}
		else // portamento
		{
			if (n<(CHIPNSFX_SIZE-3)) // SFX: ignore
				play_flag[4]=1,play_flag[5]=0;
		}
		return 1; // valid
	}
	return 0; // neither!
}

void savenote(int q,int n,int d) // bool,note,delta
{
	if (play_size)
	{
		if (past_size!=play_size)
		{
			past_size=play_size;
			if (q)
				savechar+=fprintf(f,"%s$6F+%i",savegaps(),play_size);
		}
		if (n>0&&n<CHIPNSFX_SIZE-2) // note or SFX?
		{
			if (!play_flag[4]||play_flag[5]) // ignore changes within a portamento!
			{
				if (past_flag[0]!=play_flag[0])
				{
					past_flag[0]=play_flag[0];
					if (q)
						savechar+=fprintf(f,"%s$00F8,$%02X",savegaps(),play_flag[0]);
				}
				if (((past_flag[1]!=play_flag[1])&&play_flag[0]))
				{
					past_flag[1]=play_flag[1];
					if (q)
						savechar+=fprintf(f,"%s$00FE,$%02X",savegaps(),play_flag[1]);
				}
				if (past_flag[2]!=play_flag[2])
				{
					past_flag[2]=play_flag[2];
					if (q)
						savechar+=fprintf(f,"%s$00FA,$%02X",savegaps(),play_flag[2]);
				}
				if (((past_flag[3]!=play_flag[3])&&play_flag[2]))
				{
					past_flag[3]=play_flag[3];
					if (q)
						savechar+=fprintf(f,"%s$00FF,$%02X",savegaps(),play_flag[3]);
				}
			}
			if (((past_flag[4]!=play_flag[4]||past_flag[5]!=play_flag[5])&&n<(CHIPNSFX_SIZE-3)))
			{
				past_flag[4]=play_flag[4],past_flag[5]=play_flag[5];
				if (q)
					savechar+=fprintf(f,"%s$00F%c,$%02X",savegaps(),'C'+play_flag[4],play_flag[5]);
			}
		}
		if (q)
		{
			if (n>(CHIPNSFX_SIZE-4))
				savechar+=fprintf(f,"%s$006%c",savegaps(),'F'+n-CHIPNSFX_SIZE); // SFX/gap/padding/rest
			else
			{
				n+=d; if (n>0&&n<=(CHIPNSFX_SIZE-4))
					savechar+=fprintf(f,"%s%i",savegaps(),n-1); // note
				else
					savechar+=fprintf(f,"%s$006F",savegaps()); // overflow = rest
			}
		}
	}
}
void savescore(int q,int j,int d) // boolean,score,delta
{
	int a,b,c,i,l;
	if ((i=0)>=(l=scrln[j]))
		return;
	savedefb(q);
	play_size=a=0;
	while (i<l)
	{
		if (q&&((c=save_packscore[j][i])<0)) // end loop
		{
			savenote(q,a,d);
			savechar+=fprintf(f,"%s$00F4,0",savegaps());
			play_size=a=0;
			i-=c;
		}
		else
		{
			if (c=score[j][i][0])
			{
				savenote(q,a,d);
				a=c;
				if (q&&(c=save_packscore[j][i])>0) // set loop
				{
					savechar+=fprintf(f,"%s$00F4,%i",savegaps(),c);
					// force reset if required
					if ((c=save_packflags[j][i])&(1<<7))
						past_size=256;
					for (b=0;b<6;++b)
						if (c&(1<<b))
							past_flag[b]=256;
				}
				saveplayback(a,score[j][i][1]);
				play_size=0;
			}
			play_size+=savesizes;
			++i;
		}
	}
	savenote(q,a,d); // don't peek
	savecrlf(q);
}

void savechaninit(void)
{
	MEMVOID(past_flag);
	MEMVOID(play_flag);
	past_flag[0]=play_flag[0]=255; // exception: initial volume is always max
	past_pose=play_pose=0;
	past_size=play_size=1;
}
#define SET_AS_DIRTY(x,y,q) (x=((q)&&((x==y)||(x<0)))?y:256)
int savechanprev(int n)
{
	int c,i,j=-1,k;
	if (!(n&256))
		for (c=0;c<3;++c)
			for (i=0;i<orders;++i)
				if (order[c][i][0]==n)
				{
					if (i==0||i==loopto)
						return 256;
					if (order[c][i][1]!=order[c][i-1][1])
						return 256;
					if (j==(k=order[c][i-1][0])||j<0)
						j=k;
					else
						return 256;
				}
	return j;
}
int savechannext(int p)
{
	int c,i,j=-1,k;
	if (!(p&256))
		for (c=0;c<3;++c)
			for (i=0;i<orders;++i)
				if (order[c][i][0]==p)
				{
					if (i==loopto-1||i==orders-1)
						return 256;
					if (order[c][i][1]!=order[c][i+1][1])
						return 256;
					if (j==(k=order[c][i+1][0])||j<0)
						j=k;
					else
						return 256;
				}
	return j;
}
void savechanexit(int q)
{
	int c,i,j,k;
	for (i=0;i<256;++i)
		for (c=0;c<3;++c)
		{
			j=order[c][i][0];
			if (save_chan[j]==q&&save_hits[j]>1)
			{
				if (((k=savechanprev(j))&256)||(k!=order[c][i-1][0])||savechannext(k)!=j)
					fprintf(f,"%s%c%02x:\n",title0,'a'+q,j);
				past_size=save_size[j];
				MEMLOAD(past_flag,save_flag[j]);
				savescore(1,j,0);
				if (((k=savechannext(j))&256)||(k!=order[c][i+1][0])||savechanprev(k)!=j)
					fprintf(f," db $00F1\n");
				save_chan[j]=-1; // don't touch it ever again
			}
		}
}

void savepackorderz(int q,int c,int i,int l)
{
	while (i<l)
	{
		int j=i;
		int bestoffset,bestlength,bestsaving=0;
		while (++j<l)
		{
			int k=0,m;
			while (((j+k)<l)&&(m=order[c][i+k][0])==order[c][j+k][0]&&order[c][i+k][1]==order[c][j+k][1]&&!(q&&save_pack[m]))
				++k;
			k-=k%(j-i); // align loop
			int saving=0;
			for (m=i;m<j;++m)
				saving+=save_tiny[order[c][m][0]]?1:3;
			saving=(saving*k/(j-i))-4;
			if ((bestsaving<saving)) // check whether we actually save any data
				bestsaving=saving,bestlength=k,bestoffset=j-i;
		}
		if (bestsaving)
		{
			save_packorder[c][i]=1+(bestlength/bestoffset);
			for (j=0;j<bestoffset;++j)
				save_pack[order[c][i++][0]]|=!q; // avoid using this pattern again
			save_packorder[c][i]=-bestlength;
			for (j=0;j<bestlength;++j)
				save_pack[order[c][i++][0]]|=!q; // ...if there's a later time!
		}
		else
			++i;
	}
}
void savepackorders(int q) // `q` is NONZERO if this is the second pass that must avoid clashing with the first
{
	int c;
	for (c=0;c<3;++c)
		if (loopto<0)
			savepackorderz(q,c,0,save_full[c]);
		else
			savepackorderz(q,c,0,loopto),savepackorderz(q,c,loopto,save_full[c]);

}
void savepackscores(int q) // `q` is NONZERO if this is the second pass that must avoid clashing with the first
{
	int n,i,j,k,l;
	for (n=0;n<256;++n)
		if ((l=scrln[n])>1&&!(q&&save_pack[n]))
		{
			past_size=save_size[n];
			MEMLOAD(past_flag,save_flag[n]);
			i=0;
			while (i<l)
			{
				int bestsaving=0,bestoffset,bestlength,best_size,best_flag[6];
				if (score[n][i][0])
				{
					j=i;
					while (++j<l)
					{
						k=0;
						while (((j+k)<l)&&score[n][i+k][0]==score[n][j+k][0]&&score[n][i+k][1]==score[n][j+k][1])
							++k;
						// align loop to a valid length
						k-=(k%(j-i));
						if (j+k<l)
							while (k>0&&!score[n][j+k][0])
								k-=(j-i);
						if (k>0)
						{
							// examine whether the loop is profitable
							int m,o,saving=0,temp_size=past_size,temp_flag[6];
							MEMLOAD(temp_flag,past_flag);
							MEMLOAD(play_flag,past_flag);
							play_size=0;
							for (m=i;m<j;++m)
							{
								if (saveplayback(score[n][m][0],score[n][m][1]))
								{
									++saving; // note
									if (play_size&&(temp_size!=play_size))
										temp_size=play_size,++saving; // warning, size can be delayed!
									for (o=0;o<4;++o)
										if (temp_flag[o]!=play_flag[o])
											temp_flag[o]=play_flag[o],saving+=2; // flag
									if (temp_flag[4]!=play_flag[4]||temp_flag[5]!=play_flag[5])
										temp_flag[4]=play_flag[4],temp_flag[5]=play_flag[5],saving+=2; // 00FC/00FD
									play_size=0;
								}
								play_size+=savesizes;
							}
							if (temp_size!=play_size)
								++saving; // delayed size
							if (past_size!=play_size)
								--saving; // current size
							for (o=0;o<4;++o)
								if (past_flag[o]!=play_flag[o])
									saving-=2; // flag
							if (past_flag[4]!=play_flag[4]||past_flag[5]!=play_flag[5])
								saving-=2; // 00FC/00FD
							saving=(saving*k/(j-i))-4;
							if (saving>bestsaving)
								bestsaving=saving,bestlength=k,bestoffset=j-i,best_size=play_size,MEMLOAD(best_flag,play_flag);
						}
					}
				}
				if (bestsaving)
				{
					save_pack[n]|=!q; // avoid using this pattern again
					// store which flags must be reset
					k=(best_size!=past_size&&past_size>=0)?(1<<7):0;
					for (j=0;j<6;++j)
						if (best_flag[j]!=past_flag[j]&&past_flag[j]>=0)
							k|=1<<j;
					save_packflags[n][i]=k;
					save_packscore[n][i]=1+(bestlength/bestoffset);
					i+=bestoffset;
					save_packscore[n][i]=-bestlength;
					i+=bestlength;
					past_size=best_size;
					MEMLOAD(past_flag,best_flag);
				}
				else
				{
					if (saveplayback(score[n][i][0],score[n][i][1]))
						MEMLOAD(past_flag,play_flag),past_size=0;
					past_size+=savesizes;
					++i;
				}
			}
		}
}
int flag_ll=0; // include labels as header
void savedataprepare(void) // insert dummy padding
{
	int a,b,c,i,j,k,l;
	for (i=0;i<256;++i)
		if (k=scrln[i])
		{
			if (!score[i][0][0])
				score[i][0][0]=CHIPNSFX_SIZE-1;
			do
			{
				l=a=b=0,c=j=1;
				while (j<=k)
				{
					if (score[i][j][0]||j==k)
					{
						if ((b==c)&&(c*2==a)&&(a+b+c==j))
							++l,score[i][j-b-a][0]=CHIPNSFX_SIZE-1; // aabc -> aAbc
						else if ((c==a)&&(a*2==b))
							++l,score[i][j-b][0]=CHIPNSFX_SIZE-1,b=c; // abbc -> abBc
						else if ((j==k)&&(a==b)&&(b*2==c))
							++l,score[i][j-a][0]=CHIPNSFX_SIZE-1,c=b; // abcc -> abcC
						a=b,b=c,c=0;
					}
					++c,++j;
				}
			}
			while (l); // iteration is important!
		}
}
void savedatacleanup(void) // remove padding
{
	int i,j,k;
	for (i=0;i<256;++i)
		if (k=scrln[i])
			for (j=0;j<k;++j)
				if (score[i][j][0]==CHIPNSFX_SIZE-1)
					score[i][j][0]=0;
}
int savedata(char *s) // exports the current song as an INCLUDE file; 0 is ERROR
{
	int a,b,c,i=256,j,k;
	savedataprepare();
	// -t: check whether timings can be abridged
	savesizes=1;
	if (flag_t&&divisor>1) // ensure i=256 beforehand
	{
		for (i=0;i<256;++i)
			if (((k=scrln[i])*divisor)>128)
			{
				// perform a deeper analysis
				j=a=0;
				while (j<k)
				{
					if (score[i][j++][0])
					{
						if (a>128)
							break;
						else
							a=0;
					}
					a+=divisor;
				}
				if (a>128)
					break;
			}
		if (i>=256)
			savesizes=divisor; // timings can be simplified
	}
	if ((flag_t>1&&i<256)||!s||!(f=(strcmp(s,"-")?fopen(s,flag_a?"a":"w"):stdout)))
		return savedatacleanup(),0;
	//if (!flag_q)
		fprintf(f,"; " MY_FILE " %i Hz\n; %s\n; %s\n; %s",hertz,title1,title2,chipnsfxlf);
	if (flag_ll)
		fprintf(f,"%sa_z:\n dw %sa-$-2\n dw %sb-$-2\n dw %sc-$-2\n",title0,title0,title0,title0);
	// simplify loops: ABC.LOOP.XYZXYZXYZ -> ABC.LOOP.XYZ
	for (c=0;c<3;++c)
		if (loopto>=0)
		{
			k=0;
			while (++k<orders-loopto)
			{
				if (!((orders-loopto)%k))
				{
					for (i=loopto+k;i<orders;++i)
						if ((order[c][i][0]!=order[c][i-k][0])||(order[c][i][1]!=order[c][i-k][1]))
							break;
					if (i>=orders)
						break;
				}
			}
			save_full[c]=loopto+k;
		}
		else
			save_full[c]=orders;
	// check for empty tracks
	for (c=0;c<3;++c)
	{
		for (i=0;i<save_full[c];++i)
		{
			j=order[c][i][0];
			for (k=0;k<scrln[j];++k)
				if ((a=score[j][k][0])&&(a<(CHIPNSFX_SIZE-2)))
					break;
			if (k<scrln[j])
				break;
		}
		if (i>=save_full[c])
			save_full[c]=0;
	}
	// perform preliminary playback: build flags and statistics
	MEMFULL(save_flag);
	MEMFULL(save_size);
	MEMFULL(save_chan);
	MEMVOID(save_hits);
	MEMVOID(save_tiny);
	MEMVOID(save_pack);
	MEMVOID(save_packscore);
	MEMVOID(save_packorder);
	MEMVOID(save_packflags);
	for (c=0;c<3;++c)
	{
		savechaninit();
		for (i=0;i<save_full[c];++i)
		{
			j=order[c][i][0];
			SET_AS_DIRTY(save_size[j],play_size,1);
			for (a=0;a<6;++a)
				SET_AS_DIRTY(save_flag[j][a],play_flag[a],1);
			savescore(0,j,0); // update flags
		}
		if (loopto>=0) // ensure consistency around the looping extremes
		{
			for (i=loopto;i<save_full[c];++i) // special case: look for tracks that loop before playing anything
			{
				j=order[c][i][0];
				SET_AS_DIRTY(save_size[j],play_size,1);
				for (a=0;a<6;++a)
					SET_AS_DIRTY(save_flag[j][a],play_flag[a],1);
				savescore(0,j,0); // update flags again
			}
		}
	}
	for (i=0;i<256;++i)
	{
		j=scrln[i]-1;
		while (j>0&&(!(k=score[i][j][0]))) // ..||k>=(CHIPNSFX_SIZE-0)))
			--j;
		save_tiny[i]=(j<=0); // pattern has <2 notes
	}
	// pack all loops in the song and the patterns
	switch (flag_r)
	{
		case +0: // ---
			savepackscores(0);
			savepackorders(1);
			break;
		case +2: // -r
			savepackorders(0);
			savepackscores(1);
			break;
		case +3: // -rR
			savepackorders(0);
			savepackscores(0);
			break;
		//case +1: // -R
	}
	// assign primary, secondary and tertiary pattern categories
	for (c=0;c<3;++c)
		for (i=0;i<save_full[c];)
		{
			if ((j=save_packorder[c][i])<0)
				i-=j; // compressed repetitions become simplified
			else
			{
				++save_hits[j=order[c][i][0]]; // secondary and tertiary patterns have >1 here
				if ((k=save_chan[j])!=c)
					save_chan[j]=k<0?c:3; // tertiary patterns are tagged "3"
					++i;
			}
		}
	for (i=0;i<256;++i)
		if (save_tiny[i]&&save_hits[i]>1)
			if ((savechanprev(i)&256)&&(savechannext(i)&256)) // not in a chain!
				save_hits[i]=1; // lone patterns with just a note become primary
	// export the song's three channels
	for (b=0;b<3;++b)
	{
		c=flag_bb[b];
		savechaninit();
		// channel header
		fprintf(f,"%s%c:\n",title0,'a'+c);
		if (savesizes<divisor&&save_full[c])
			fprintf(f," db $00F5,%i\n",divisor);
		// export the channel's primary patterns
		for (i=0;i<save_full[c];)
		{
			if ((j=save_packorder[c][i])<0)
			{
				fprintf(f," db $00F4,0\n");
				i-=j;
			}
			else
			{
				play_pose=transpose+(signed char)order[c][i][1];
				if (i==loopto)
				{
					fprintf(f,"%s%c_l:\n",title0,'a'+c);
					k=save_full[c];
					while (--k>i&&save_tiny[order[c][k][0]])
						{}
					j=transpose+(signed char)order[c][k][1];
					if (past_pose!=j)
						past_pose=256;
					past_size=save_size[j=order[c][i][0]];
					MEMLOAD(past_flag,save_flag[j]);
				}
				if ((j=save_packorder[c][i])>0)
				{
					fprintf(f," db $00F4,%i\n",j);
					k=i;
					while (save_packorder[c][++k]>=0)
						{}
					while (--k>i&&save_tiny[order[c][k][0]])
						{}
					j=transpose+(signed char)order[c][k][1];
					if (past_pose!=j)
						past_pose=256;
					past_size=save_size[j=order[c][i][0]];
					MEMLOAD(past_flag,save_flag[j]);
				}
				j=order[c][i][0];
				if (past_pose!=play_pose)
					if (past_pose==256||!save_tiny[j]||save_hits[j]>1)
						fprintf(f," db $00F6,%+i\n",past_pose=play_pose);
				if (save_hits[j]>1)
				{
					k=savechanprev(j);
					if ((k&256)||k!=order[c][i-1][0]||savechannext(k)!=j)
					{
						k=save_chan[j];
						a=!(flag_b&(1<<(c+(k<3?0:3))));
						fprintf(f," db $00F%c\n  d%c %s%c%02x-$-%c\n",a?'2':'3',a?'w':'b',title0,'a'+k,j,'1'+a);
					}
					savescore(0,j,0); // update flags
				}
				else
					savescore(1,j,play_pose-past_pose);
				++i;
			}
		}
		// channel footer
		if (loopto<0||!i)
			savechar+=fprintf(f," db $00F0\n");
		else
			savechar+=fprintf(f," db $00F2\n  dw %s%c_l-$-2\n",title0,'a'+c);
		// export the channel's secondary patterns
		savechanexit(c);
	}
	// export the song's tertiary patterns
	savechanexit(3);
	//if (!flag_q)
		fprintf(f,"; %s",chipnsfxlf);
	if (f!=stdout)
		fclose(f);
	return savedatacleanup(),1;
}

/*--------*/

int activepanel=0,activeoctave=4,activeparameter=0;
int activescore=0,activescoreitem=0,activeinstr=1,activeinstritem=0,activeorder=0,activeorderitem=0;
int selectscore=0,selectscoreitem=0,selectorder=0,selectorderitem=0; // to use with SHIFT+cursors

int dirty,can_undo,can_redo,flag_uu=0;
void inc_dirty(void)
{
	dirty=dirty<1?1:2; // 0 -> 1, 1 -> 2, 2 -> 2
}
void dec_dirty(void)
{
	dirty=dirty>1?2:0; // 0 -> 0, 1 -> 0, 2 -> 2
}
void undo_set(void)
{
	loopto_=loopto;
	orders_=orders;
	instrs_=instrs;
	MEMLOAD(order_,order);
	MEMLOAD(score_,score);
	MEMLOAD(scrln_,scrln);
	MEMLOAD(instr_,instr);
	inc_dirty();
	can_undo=1;
	can_redo=0;
}
INLINE void intswp(int *i,int *j)
{
	int k=*i; *i=*j; *j=k;
}
void *memswp(void *t,void *s,int l)
{
	char a,*r=(char *)t,*q=(char *)s;
	while (l--)
		a=*r,*r++=*q,*q++=a;
	return t;
}
void undo_get(int redo)
{
	if ((!flag_uu&&can_undo)||((redo&&can_redo)||(!redo&&can_undo)))
	{
		intswp(&loopto,&loopto_);
		intswp(&orders,&orders_);
		intswp(&instrs,&instrs_);
		memswp(order,order_,sizeof(order));
		memswp(score,score_,sizeof(score));
		memswp(scrln,scrln_,sizeof(scrln));
		memswp(instr,instr_,sizeof(instr));
		if (!flag_uu)
		{
			if (dirty<2)
				dirty=1-dirty; // 0 -> 1, 1 -> 0, 2 ->
		}
		else
		{
			can_redo=!(can_undo=redo);
			redo?inc_dirty():dec_dirty();
		}
	}
}

/*--------*/

// both must be N*100
#define SND_HIGHEST 48000
int SND_QUALITY=44100;
int SND_QUANTUM;
int SND_DEPTH=1,SND_DEPTHLEFT=0,SND_DEPTHBITS;

#define SND_FIXED 8
static const int snd_freqs[]={(0x3BB907),(0x385EEB),(0x3534F9),(0x32387D),(0x2F66E9),(0x2CBDD4),
	(0x2A3AF9),(0x27DC33),(0x259F7C),(0x2382E9),(0x2184AD),(0x1FA314)}; // C-0..B#0 @ 4<<SND_FIXED MHz
static unsigned char snd_ampls[16]={0,1,1,1,2,3,4,5,8,11,15,21,30,43,60,85};
	//85,60,43,30,21,15,11,8,5,4,3,2,1,1,1,0 }; // second half for triangular waves
int snd_playing=0,snd_looping=0,snd_order=0,snd_score=0,snd_testn,snd_testd,snd_testi=0,
	snd_ticks=0,snd_count=0,snd_noise=0,snd_noises=0,snd_noiser=0,snd_noisez=0,snd_noised=0;
int snd_bool[3]={1,1,1},snd_note[3],snd_freq[3],snd_freq0[3],snd_ampl[3],snd_ampl0[3],snd_ampld[3],
	snd_brake[3],snd_noisy[3],snd_nois0[3],snd_noisd[3],snd_sfx[3],snd_sfx0[3],snd_sfxd[3],snd_freqz[3];

int snd_notefreq(int i)
{
	if (i<0||i>(CHIPNSFX_SIZE-4))
		return 0;
	return (((snd_freqs[i%12])>>(i/12))+1)>>1;
}

#define SND_BUFFERS (1<<2)
#define SND_BUFFERX (1<<2)

unsigned char snd_buffer[SND_BUFFERS][((SND_HIGHEST+50)/100)*SND_BUFFERX];

void snd_stop(void)
{
	snd_playing=0;
	snd_ampl[0]=snd_ampl[1]=snd_ampl[2]=0;
}
void snd_play(int m) // 0=normal,!0=pattern
{
	if (orders&&scrln[order[0][0][0]]&&scrln[order[1][0][0]]&&scrln[order[2][0][0]])
	{
		int c;
		for (snd_noisez=snd_noised=c=0;c<3;++c)
		{
			snd_freq[c]=snd_ampl[c]=snd_noisy[c]=snd_sfx[c]=
				snd_freq0[c]=snd_freqz[c]=snd_ampld[c]=
				snd_brake[c]=snd_nois0[c]=snd_noisd[c]=
				snd_sfx0[c]=snd_sfxd[c]=0;
			snd_note[c]=snd_ampl0[c]=255;
		}
		snd_count=snd_ticks=snd_noise=snd_noises=snd_noiser=0;
		snd_looping=m;
		snd_playing=1;
	}
}

int snd_playnote(int c,int n,int d,int i) // channel, note, transposition, instrument; 1 if it extends the current note, 0 if new
{
	if (n>0)
	{
		if (n==(CHIPNSFX_SIZE-2))
			return snd_brake[c]=!snd_brake[c],1; // brake
		else if (n<(CHIPNSFX_SIZE-0))
		{
			if (n>(CHIPNSFX_SIZE-4))
			{
				snd_note[c]=(CHIPNSFX_SIZE-4);
				n=0; // sfx
			}
			else
			{
				snd_note[c]=(n+=transpose+d-1);
				n=snd_notefreq(n); // note
			}
			snd_sfx[c]=0;
			if (i) // instrument
			{
				snd_freq0[c]=n;
				snd_ampl[c]=snd_ampl0[c]=instr[i][0];
				snd_ampld[c]=instr[i][1];
				snd_brake[c]=0;
				if (n=snd_noisy[c]=snd_nois0[c]=instr[i][2])
				{
					snd_noisez=n;
					if ((n=snd_noisd[c]=instr[i][3])!=0x80)
						snd_noised=n;
				}
				snd_sfx0[c]=instr[i][4];
				snd_sfxd[c]=instr[i][5];
			}
			else // portamento
				return snd_sfxd[c]=n,snd_sfx0[c]=-1;
		}
		else
			snd_ampl[c]=0; // rest
		return 0;
	}
	return 1;
}
void snd_playchan(int c)
{
	int i,j;
	if (snd_ampl[c])
	{
		if (!snd_freq0[c])
			snd_freq[c]=0; // allow pure noise
		if (!snd_brake[c])
		{
			i=(unsigned char)snd_ampld[c];
			if (i==0x60||i==0x80) // zero depth: future special cases for hardware envelopes!?
			{}
			else if (i>=0x60&&i<0xA0)
			{
				int j=(signed char)(i<<3);
				if (i&0x80) // timbre or tremolo
					if (snd_ampl[c]!=snd_ampl0[c])
						j=0;
				snd_ampl[c]=snd_ampl0[c]+j;
				keepwithin(&snd_ampl[c],0,255);
			}
			else
			{
				// basic amplitude envelope
				if ((snd_ampl[c]+=(signed char)i)<0)
					snd_ampl[c]=0;
				else if (snd_ampl[c]>255)
					snd_ampl[c]=255;
			}
		}
		if (snd_noisd[c]==0x80) // special case: instant noise
			snd_noisy[c]=0;
		if (snd_freq0[c]&&snd_sfxd[c])
		{
			if (!snd_sfx0[c]) // arpeggio
			{
				if ((snd_sfx[c]+=0x40)>((i=(snd_sfxd[c]&0x0F))?0x80:0x40))
					snd_sfx[c]=0;
				if (snd_sfx[c]&0x80)
					{}
				else if (snd_sfx[c]&0x40)
					i=snd_sfxd[c]>>4;
				else
					i=0;
				// special cases
				if (i==13)
					i=24; // D=24
				else if (i>13)
				{
					i=(i&1)?-12:-24; // F=-12, E=-24
					snd_sfx[c]=0x40; // lock arpeggio
				}
				snd_freq0[c]=snd_notefreq(snd_note[c]+i);
			}
			else if (snd_sfx0[c]>0) // vibrato et al.
			{
				j=snd_freq0[c]>>(8+1+SND_FIXED); // proportional v.
				if (!(i=(snd_sfxd[c]&7))) // slow glissando?
					i=1;
				i=(j+1)<<(i-1);
				if (snd_sfxd[c]&8)
					i=-i; // signed vibrato
				if (j=(snd_sfxd[c]&0xF0))
				{
					if (((snd_sfx[c]+=16)&0xF0)==j)
					{
						if (j=(snd_sfxd[c]&0x07)) // vibrato?
						{
							if (2&(snd_sfx[c]=((snd_sfx[c]+1)&3))) // +I -I -I +I ...
							//if (!(snd_sfx[c]=((snd_sfx[c]+1)&1))) // ultra soft vibrato
								i=-i;
						}
						else
							snd_sfx[c]=0; // slow glissando
						snd_freq0[c]-=i<<(1+SND_FIXED);
					}
				}
				else // glissando
					snd_freq0[c]-=i<<(1+SND_FIXED);
				if (snd_freq0[c]&(1<<(8+6+SND_FIXED)))
					snd_freq0[c]=0; // safety!
			}
			else // portamento
				snd_freq0[c]=(snd_sfxd[c]+snd_freq0[c]+(snd_sfxd[c]>snd_freq0[c]))/2;
		}
	}
}

void snd_frame(int h)
{
	int c,i,q=3;
	if (snd_count<=0)
	{
		if (snd_noisez)
		{
			if ((i=snd_noised)==0x80)
				{}
			else
			{
				snd_noisez+=(signed char)i;
				keepwithin(&snd_noisez,1,255);
			}
		}
		if (snd_ticks<=0)
		{
			if (snd_order>=orders)
				if (snd_playing&&(loopto<0||snd_looping||!scrln[order[0][snd_order=loopto][0]]||(h&&!--flag_n)))
					snd_stop();
			if (snd_playing)
			{
				for (q=c=0;c<3;++c) // walk thru the scores
				{
					i=order[c][snd_order][0];
					if (snd_playnote(c,score[i][snd_score][0],(signed char)order[c][snd_order][1],score[i][snd_score][1]))
						snd_playchan(c);
				}
				if ((++snd_score)>=scrln[i])
				{
					snd_score=0;
					if (!snd_looping)
						++snd_order;
				}
				snd_ticks+=divisor;
			}
		}
		for (c=0;c<q;++c)
			snd_playchan(c);
		--snd_ticks;
		snd_count+=100;
		for (c=0;c<3;++c)
			snd_freqz[c]=(snd_freq0[c]*(4+(flag_zz&3)))/4;
	}
	snd_count-=hertz;
}

int snd_wavez=0; // self-adjusting step
void snd_waves(unsigned char *t)
{
	int c,i,j=SND_QUANTUM,k,l,m,n=snd_noisez,z;
	if (!(n>>=2)) // 2 for CPC, 3 for ST
		++n;
	n<<=SND_FIXED+1;
	while (j--)
	{
		k=(snd_wavez+=(125000<<(1+SND_FIXED-SND_DEPTH)))/SND_QUALITY; // adjust!
		snd_wavez%=SND_QUALITY;
		z=SND_DEPTHLEFT;
		i=1<<SND_DEPTH;
		while (i--)
		{
			for (c=0;c<3;++c)
				if (snd_ampl[c]&&(!snd_playing||snd_bool[c]))
				{
					snd_freq[c]=(m=(l=snd_freq0[c])?(snd_freq[c]+k)%(2*l):0);
					if (snd_noises||!snd_noisy[c])
						if (!l||m<snd_freqz[c])
							z+=snd_ampls[snd_ampl[c]>>4];
				}
			if ((snd_noise+=k)>=n)
			{
				snd_noiser+=snd_noises+snd_noise-j; // random seed
				snd_noises=(!(snd_noiser&1))^(!(snd_noiser&2))^(!(snd_noiser&4))^
					(!(snd_noiser&16))^(!(snd_noiser&256))^(!(snd_noiser&65536));
				snd_noise%=n;
			}
		}
		SND_DEPTHLEFT=z&SND_DEPTHBITS;
		*t++=z>>SND_DEPTH;
	}
}

#ifndef DEBUG

HWAVEOUT hWave=0;
WAVEFORMATEX fWave;
WAVEHDR pWave[SND_BUFFERS];
int iWave;

int snd_callback_last=0;
void CALLBACK snd_callback(HWAVEOUT hWo,UINT uMsg,DWORD dwInstance,DWORD dwParam1,DWORD dwParam2)
{
	//if (uMsg==MM_WOM_CLOSE)
		//hWave=0; // this should never happen!
	if (uMsg!=MM_WOM_DONE||!hWave)
		return;
	int i;
	//waveOutUnprepareHeader(hWave,&pWave[iWave],sizeof(WAVEHDR));
	for (i=0;i<SND_BUFFERX;++i)
	{
		snd_frame(!hWave);
		if (snd_testi&&snd_count<=0) // must happen after SND_FRAME!
		{
			if (!snd_playing)
				snd_playnote(1,snd_testn,snd_testd,snd_testi);
			snd_testi=0;
		}
		snd_waves(&snd_buffer[iWave][i*SND_QUANTUM]);
	}
	unsigned char *c=snd_buffer[iWave];
	i=SND_BUFFERX*SND_QUANTUM;
	int o=snd_callback_last;
	while (i--)
		{ int n=*c; *c++=128-(n+o+2*(n>o))/4; o=n; }
	snd_callback_last=o;
	//waveOutPrepareHeader(hWave,&pWave[iWave],sizeof(WAVEHDR));
	waveOutWrite(hWave,&pWave[iWave],sizeof(WAVEHDR));
	iWave=(iWave+1)%SND_BUFFERS;
}
void snd_open(void)
{
	fWave.wFormatTag=WAVE_FORMAT_PCM, fWave.cbSize=iWave=0;
	fWave.wBitsPerSample=8, fWave.nChannels=fWave.nBlockAlign=1; // 8-bit mono:
	fWave.nSamplesPerSec=fWave.nAvgBytesPerSec=SND_QUALITY; // 1 sample = 1 byte
	if (flag_m||waveOutOpen(&hWave,-1,&fWave,(DWORD)snd_callback,0,CALLBACK_FUNCTION))
		hWave=0;
	else
	{
		snd_stop();
		waveOutPause(hWave);
		int i;
		for (i=0;i<SND_BUFFERS;++i)
		{
			memset(&pWave[i],0,sizeof(WAVEHDR));
			memset(snd_buffer[i],128,SND_BUFFERX*SND_QUANTUM);
			pWave[i].lpData=(char *)&snd_buffer[i];
			pWave[i].dwBufferLength=SND_BUFFERX*SND_QUANTUM;
			waveOutPrepareHeader(hWave,&pWave[i],sizeof(WAVEHDR));
			waveOutWrite(hWave,&pWave[i],sizeof(WAVEHDR));
		}
		waveOutRestart(hWave);
	}
}
void snd_close(void)
{
	if (hWave)
	{
		//snd_stop();
		HWAVEOUT h=hWave;
		hWave=0;
		waveOutReset(h);
		int i;
		for (i=0;i<SND_BUFFERS;++i)
			waveOutUnprepareHeader(h,&pWave[i],sizeof(WAVEHDR));
		waveOutClose(h);
	}
}

#endif

#define fputcc(x,y) { fputc((x),(y)); fputc((x)>>8,(y)); }
#define fputcccc(x,y) { fputcc((x),(y)); fputcc((x)>>16,(y)); }
int savewave(char *s,int i) // exports the current song as a WAVE file; 0 is ERROR
{
	int l;
	if (!s||!(f=fopen(s,"wb")))
		return 0;
	fwrite("RIFFFFFFWAVEfmt \x10\x00\x00\x00\x01\x00\x01\x00",1,24,f);
	fputcccc(SND_QUALITY,f);
	fputcccc(SND_QUALITY,f);
	fwrite("\x01\x00\x08\x00""dataaaaa",1,12,f);
	if (i<0)
		i=0; // beware of non-looping songs!
	snd_order=i;
	//loopto=-1; // forbid looping
	snd_play(0);
	l=0;
	flag_q_init(orders);
	for (;;)
	{
		flag_q_loop(snd_order);
		snd_frame(!0);
		if (!snd_playing)
			break;
		snd_waves(t);
		l+=fwrite(t,1,SND_QUANTUM,f);
	}
	flag_q_exit(s);
	fseek(f,40,SEEK_SET);
	fputcccc(l,f);
	l+=36;
	fseek(f,4,SEEK_SET);
	fputcccc(l,f);
	fclose(f);
	if (flag_v)
		printf("WAVE: %i bytes.\n",l+8);
	return 1;
}

int saveym3b(char *s,int i) // exports the current song as an YM3b file; 0 is ERROR
{
	int l,o=0;
	// measure song in advance
	if (i<0)
		i=0; // beware of non-looping songs!
	snd_order=i;
	l=0;
	while (i<orders)
	{
		if (i==loopto)
			o=l;
		l+=scrln[order[0][i++][0]];
	}
	l*=divisor;
	o*=divisor;
	if (hertz<50)
	{
		l<<=1;
		o<<=1;
	}
	else if (hertz>50)
	{
		l>>=1;
		o>>=1;
	}
	char *y=(char *)malloc(l*14);
	if (!y)
		return 0;
	if (!s||!(f=fopen(s,"wb")))
	{
		free(y);
		return 0;
	}
	fwrite(o?"YM3b":"YM3!",1,4,f);
	//loopto=-1; // forbid looping
	snd_play(i=0);
	flag_q_init(l);
	while (i<l)
	{
		flag_q_loop(i);
		snd_frame(!0);
		snd_frame(!0); // reduce 100 Hz to 50 Hz
		int j;
		y[i     ]=j=(snd_freq0[0]>>SND_FIXED); // YM3b @ 2 MHz
		y[i+l   ]=j>>8;
		y[i+l* 2]=j=(snd_freq0[1]>>SND_FIXED);
		y[i+l* 3]=j>>8;
		y[i+l* 4]=j=(snd_freq0[2]>>SND_FIXED);
		y[i+l* 5]=j>>8;
		y[i+l* 6]=snd_noisez>>3; // 0..31
		y[i+l* 7]=(snd_noisy[0]?0:8)|(snd_noisy[1]?0:16)|(snd_noisy[2]?0:32)|((snd_freq0[0]&&snd_ampl[0])?0:1)|((snd_freq0[1]&&snd_ampl[1])?0:2)|((snd_freq0[2]&&snd_ampl[2])?0:4);
		y[i+l* 8]=snd_ampl[0]>>4; // 0..15
		y[i+l* 9]=snd_ampl[1]>>4;
		y[i+l*10]=snd_ampl[2]>>4;
		y[i+l*11]=y[i+l*12]=0;
		y[i+l*13]=-1; // IDLE
		++i;
	}
	flag_q_exit(s);
	fwrite(y,1,l*14,f);
	if (o)
		fputcccc(o,f);
	fclose(f);
	free(y);
	if (flag_v)
		printf("YM3b: %i bytes.\n",l+8);
	return 1;
}

/*--------*/

#ifndef DEBUG

int GUI_HEIGHT;//=25
int GUI_WIDTH;//=80
#define GUI_ESCAPE VK_ESCAPE
#define GUI_F1 VK_F1
#define GUI_F2 VK_F2
#define GUI_F3 VK_F3
#define GUI_F4 VK_F4
#define GUI_F5 VK_F5
#define GUI_F6 VK_F6
#define GUI_F7 VK_F7
#define GUI_F8 VK_F8
#define GUI_F9 VK_F9
#define GUI_F11 VK_F11
#define GUI_F12 VK_F12
#define GUI_UP VK_UP
#define GUI_DOWN VK_DOWN
#define GUI_LEFT VK_LEFT
#define GUI_RIGHT VK_RIGHT
#define GUI_PRIOR VK_PRIOR
#define GUI_NEXT VK_NEXT
#define GUI_HOME VK_HOME
#define GUI_END VK_END
#define GUI_INSERT VK_INSERT
#define GUI_DELETE VK_DELETE
#define GUI_ALT (LEFT_ALT_PRESSED|RIGHT_ALT_PRESSED)
#define GUI_SHIFT SHIFT_PRESSED
#define GUI_CONTROL (LEFT_CTRL_PRESSED|RIGHT_CTRL_PRESSED)
#define GUI_ALTF4() (key_code==GUI_F4&&(key_flag&GUI_ALT))

int key_code,key_char,key_flag;
unsigned char gui_buffer[MAXSTRLEN];

DWORD iTerminalMode,iKeyboardMode;
HANDLE hTerminal,hKeyboard;
INPUT_RECORD pBuffer;
CONSOLE_CURSOR_INFO pConsoleCursorInfo;
CONSOLE_SCREEN_BUFFER_INFO pConsoleScreenBufferInfo;

void gui_clrscr(int i) // clear screen
{
	COORD dwCursor={0,0}; int t; // bug on Windows 10 if NULL instead of (PDWORD)&t
	FillConsoleOutputAttribute(hTerminal,i,GUI_HEIGHT*GUI_WIDTH,dwCursor,(PDWORD)&t);
	FillConsoleOutputCharacter(hTerminal,' ',GUI_HEIGHT*GUI_WIDTH,dwCursor,(PDWORD)&t);
}
void gui_colour(int i) // colour 7 is NOT #808080!
{
	SetConsoleTextAttribute(hTerminal,0
		|(i&0x01?BACKGROUND_BLUE:0)
		|(i&0x02?BACKGROUND_RED:0)
		|(i&0x04?BACKGROUND_GREEN:0)
		|(i&0x08?BACKGROUND_INTENSITY:0)
		|(i&0x10?FOREGROUND_BLUE:0)
		|(i&0x20?FOREGROUND_RED:0)
		|(i&0x40?FOREGROUND_GREEN:0)
		|(i&0x80?FOREGROUND_INTENSITY:0));
}
void gui_locate(int x,int y) // locate cursor
{
	COORD dwCursor;
	dwCursor.X=x/*+(GUI_WIDTH-80)/2*/;
	dwCursor.Y=y;
	SetConsoleCursorPosition(hTerminal,dwCursor);
}
#define gui_output(x,y) WriteConsole(hTerminal,(x),(y),NULL,NULL)
#define gui_printf(...) sprintf(gui_buffer,__VA_ARGS__),gui_output(gui_buffer,strlen(gui_buffer))

int gui_init(int minx,int miny) // set screen
{
	GetConsoleScreenBufferInfo(hTerminal=GetStdHandle(STD_OUTPUT_HANDLE),&pConsoleScreenBufferInfo);
	GUI_WIDTH=pConsoleScreenBufferInfo.srWindow.Right-pConsoleScreenBufferInfo.srWindow.Left+1;
	GUI_HEIGHT=pConsoleScreenBufferInfo.srWindow.Bottom-pConsoleScreenBufferInfo.srWindow.Top+1;
	if (GUI_WIDTH<minx||GUI_HEIGHT<miny)
		return 0;
	GetConsoleMode(hTerminal,&iTerminalMode);
	SetConsoleMode(hTerminal,0);
	GetConsoleMode(hKeyboard=GetStdHandle(STD_INPUT_HANDLE),&iKeyboardMode);
	SetConsoleMode(hKeyboard,ENABLE_WINDOW_INPUT);
	GetConsoleCursorInfo(hTerminal,&pConsoleCursorInfo);
	pConsoleCursorInfo.bVisible=0;
	SetConsoleCursorInfo(hTerminal,&pConsoleCursorInfo);
	gui_clrscr(0);
	return 1;
}
void gui_exit(void) // reset screen
{
	gui_clrscr(pConsoleScreenBufferInfo.wAttributes);
	gui_locate(0,GUI_HEIGHT-1);
	SetConsoleTextAttribute(hTerminal,pConsoleScreenBufferInfo.wAttributes);
	SetConsoleMode(hTerminal,iTerminalMode);
	SetConsoleMode(hKeyboard,iKeyboardMode);
	pConsoleCursorInfo.bVisible=1;
	SetConsoleCursorInfo(hTerminal,&pConsoleCursorInfo);
	CloseHandle(hTerminal);
	CloseHandle(hKeyboard);
}

void gui_titles(char *s) // set window title
{
	if (!s||!*s)
		SetConsoleTitle(MY_NAME);
	else
	{
		sprintf(gui_buffer,"%s - " MY_NAME,s);
		SetConsoleTitle(gui_buffer);
	}
}

char gui_upink0,gui_upink1,gui_upink2,gui_upink3;
void gui_upinit(int i)
{
	gui_colour(gui_upink0=i?0xF0:0x80);
	gui_upink1=i?0x0F:0x08;
	gui_upink2=i?0xF9:0x81;
	gui_upink3=i?0x9F:0x18;
}
void gui_upline(int m,int n,int q)
{
	if (n)
	{
		if (q)
			gui_colour(gui_upink2);
		gui_output(&gui_buffer[0],m);
		gui_colour(q?gui_upink3:gui_upink1);
		gui_output(&gui_buffer[m],n-m);
		gui_colour(q?gui_upink2:gui_upink0);
		gui_output(&gui_buffer[n],strlen(gui_buffer)-n);
		if (q)
			gui_colour(gui_upink0);
	}
	else
		gui_output(&gui_buffer[n],strlen(gui_buffer)-n);
}
void gui_redraw(int q) // redraws parts of interface on request
{
	int i,j,k,l=GUI_HEIGHT-1,m=0,n=0; // avoid warnings; m and n are set when required
	if (q&1)
	{
		// left panel
		gui_upinit(activepanel==0);
		k=scrln[order[0][activeorder][0]];
		for (i=2;i<l;++i)
		{
			j=i-(GUI_HEIGHT+1)/2+activescore;
			gui_locate(0,i);
			if (j<(n=0)||j>=128)
				strcpy(gui_buffer," -- ------ ------ ------ ");
			else
			{
				#if 1 // hybrid: " 00..C7"
				sprintf(gui_buffer," %X%c ",j/10,48+(j%10));
				#else
				#if 1 // decimal: " 00".."127"
				sprintf(gui_buffer,"%2i%c ",j/10,48+(j%10));
				#else // hexadecimal: "00".."7F"
				sprintf(gui_buffer," %02X ",j);
				#endif
				#endif
				for (m=0;m<3;++m)
					sprintf(&gui_buffer[4+7*m],"%s ",
						musical(score[order[m][activeorder][0]][j][0],score[order[m][activeorder][0]][j][1]));
				if (j>=k)
					gui_buffer[0]=' ',gui_buffer[1]=gui_buffer[2]='/';
				if (j>=intmin(activescore,selectscore)&&j<=intmax(activescore,selectscore))
				{
					m=4+intmin(activescoreitem,selectscoreitem)*7;
					n=10+intmax(activescoreitem,selectscoreitem)*7;
				}
			}
			gui_upline(m,n,j==activescore);
		}
	}
	if (q&2)
	{
		// central panel
		gui_upinit(activepanel==1);
		for (i=2;i<l;++i)
		{
			j=i-(GUI_HEIGHT+1)/2+activeinstr;
			gui_locate(25,i);
			if (j<(n=0)||j>=256)
				strcpy(gui_buffer," -- ---------------- ----------- ");
			else if (!j)
			{
				strcpy(gui_buffer," .. << PORTAMENTO >> ........... ");
				if (j==activeinstr)
					n=(m=4)+16;
			}
			else
			{
				sprintf(gui_buffer," %02X %-16s %02X%02X%02X%02X%01X%02X ",j,
					&instr[j][8],
					instr[j][0],instr[j][1],
					instr[j][2],instr[j][3],
					instr[j][4],instr[j][5]);
				if (j>=instrs)
					gui_buffer[1]=gui_buffer[2]='/';
				if (j==activeinstr)
					n=(m=4+activeinstritem)+1;
			}
			gui_upline(m,n,n);
		}
	}
	if (q&4)
	{
		// right panel
		gui_upinit(activepanel==2);
		for (i=2;i<l;++i)
		{
			j=i-(GUI_HEIGHT+1)/2+activeorder;
			gui_locate(58,i);
			if (j<(n=0)||j>=256)
				sprintf(gui_buffer," -- ----- ----- ----- ");
			else
			{
				int a=(scrln[order[0][j][0]]==scrln[order[1][j][0]]&&scrln[order[1][j][0]]==scrln[order[2][j][0]])?' ':'!';
				sprintf(gui_buffer," %02X %02X%+03i%c%02X%+03i%c%02X%+03i ",j,
					order[0][j][0],(signed char)order[0][j][1],a,
					order[1][j][0],(signed char)order[1][j][1],a,
					order[2][j][0],(signed char)order[2][j][1]);
				if (j==loopto)
					gui_buffer[1]=gui_buffer[2]='*';
				if (j>=orders)
					gui_buffer[1]=gui_buffer[2]='/';
				if (j>=intmin(activeorder,selectorder)&&j<=intmax(activeorder,selectorder))
				{
					m=4+intmin(activeorderitem,selectorderitem)*6;
					n=9+intmax(activeorderitem,selectorderitem)*6;
				}
			}
			gui_upline(m,n,j==activeorder);
		}
	}
	if (q&8)
	{
		// status bar
		gui_locate(0,l);
		gui_colour(activepanel==3?0x0F:0xF8);
		sprintf(gui_buffer," Octave %i  %03i/%02i Hz  Transpose %+03i  %s "
			#if 1 // decimal: " 00".."127"
			"%02X.%03i"
			#else // hexadecimal: "00".."7F"
			" %02X.%02X"
			#endif
			" %c%c%c  " MY_NAME " " MY_DATE " ",
			activeoctave,hertz,divisor,transpose,hWave?(snd_playing?(snd_looping?"LOOP":"PLAY"):"STOP"):"MUTE",snd_order,snd_score,
			snd_playing?vubars[flag_xx&1][snd_ampl[0]>>4]:'-',snd_playing?vubars[flag_xx&1][snd_ampl[1]>>4]:(dirty?'*':'-'),snd_playing?vubars[flag_xx&1][snd_ampl[2]>>4]:'-');
		if (activepanel==3)
		{
			switch(activeparameter)
			{
				case 0:
					m=11;
					n=14;
					break;
				case 1:
					m=15;
					n=17;
					break;
				case 2:
					m=32;
					n=35;
					break;
			}
			gui_output(gui_buffer,m);
			gui_colour(0xF9);
			gui_output(&gui_buffer[m],n-m);
			gui_colour(0x0F);
		}
		else
			n=0;
		gui_output(&gui_buffer[n],strlen(gui_buffer)-n);
	}
	if (q&16)
	{
		// panel captions
		gui_locate(0,1);
		gui_colour(activepanel==0?0x0A:0xA2);
		gui_printf(" ##    %c      %c      %c   ",snd_bool[0]?'A':'-',snd_bool[1]?'B':'-',snd_bool[2]?'C':'-');
		gui_colour(activepanel==1?0x0C:0xC4);
		gui_printf(" ##    Instrument    AmplNoisSFX ");
		gui_colour(activepanel==2?0x09:0x91);
		gui_printf(" ##   %c     %c     %c   ",snd_bool[0]?'A':'-',snd_bool[1]?'B':'-',snd_bool[2]?'C':'-');
	}
	if (q&32)
	{
		// global caption
		gui_locate(0,0);
		gui_colour(0xF8);
		gui_printf(strlen(title1)>38?" %.36s.. ":" %-38s ",title1);
		gui_printf(strlen(title2)>38?" %.36s.. ":" %-38s ",title2);
	}
}

int gui_delays;
void gui_listen(void) // handle events
{
	do
	{
		int i=0;
		do
		{
			if (gui_delays=!gui_delays)
			{
				int k=8;
				if (flag_pp&&snd_playing)
				{
					if (activescore!=snd_score)
					{
						k|=1;
						selectscore=activescore=snd_score;
					}
					if (activeorder!=snd_order)
					{
						k|=4;
						selectorder=activeorder=snd_order;
					}
				}
				gui_redraw(k);
			}
			Sleep(1000/50); // who needs more than 50Hz?
			PeekConsoleInput(hKeyboard,&pBuffer,1,(PDWORD)&i);
		}
		while (!i);
		ReadConsoleInput(hKeyboard,&pBuffer,1,(PDWORD)&i);
	}
	while ((!(pBuffer.EventType&KEY_EVENT))||(!pBuffer.Event.KeyEvent.bKeyDown));
	key_code=pBuffer.Event.KeyEvent.wVirtualKeyCode;
	key_char=pBuffer.Event.KeyEvent.uChar.AsciiChar&0xFF;
	key_flag=pBuffer.Event.KeyEvent.dwControlKeyState;
	// simple translations
	if (key_code==VK_INSERT)
	{
		if (key_flag&GUI_SHIFT)
			key_char=(key_code='V')-64; // C-v: PASTE
		else if (key_flag&GUI_CONTROL)
			key_char=(key_code='C')-64; // C-c: COPY
	}
	else if (key_code==VK_DELETE)
	{
		if (key_flag&GUI_SHIFT)
			key_char=(key_code='X')-64; // C-x: CUT
		else if (key_flag&GUI_CONTROL)
			key_char=(key_code='Z')-64; // C-z: UNDO
	}
	else if (key_code==VK_BACK)
		key_char=8;
	else if (key_code==VK_TAB)
		key_char=key_flag&GUI_SHIFT?8:9;
}

int stack_pp; // temporarily disable PLAYBACK FOLLOW!
void push_pp(void) { stack_pp=flag_pp; flag_pp=0; }
void pop_pp(void) { flag_pp=stack_pp; }

int gui_readme(char *s) // show text
{
	int l,lx=0,ly=1,x=0,y;
	char c,*r=s;
	while (*r)
	{
		if ((c=*r++)=='\n')
		{
			if (x>lx)
				lx=x;
			x=0;
			++ly;
		}
		else
		{
			if (c=='\t')
				x|=7;
			++x;
		}
	}
	if (x>lx)
		lx=x;
	gui_colour(0x0F);
	lx+=2;
	y=(GUI_HEIGHT-ly)/2;
	x=(80-lx)/2;
	ly+=y;
	r=s;
	while (y<ly)
	{
		gui_locate(x,y);
		char *t=gui_buffer;
		*t++=' ';
		l=0;
		while ((c=*r++)&&(c!='\n'))
			if (c=='\t')
				do
				{
					*t++=' ';
				} while (++l&7);

			else
				++l,*t++=c;
		*t=0;
		strpad(gui_buffer,lx);
		gui_output(gui_buffer,lx);
		++y;
	}
	for (;;)
	{
		push_pp(); gui_listen(); pop_pp();
		if (key_char==13||key_code==GUI_ESCAPE||GUI_ALTF4())
			break;
	}
	return -1; // for gui_redraw()
}
#define gui_readmef(...) ( sprintf(t,__VA_ARGS__), gui_readme(t) )

int gui_really(char *m) // yes or no: ZERO is NO
{
	int i=strlen(m)+23;
	gui_colour(0x0F);
	gui_locate((80-i)/2,GUI_HEIGHT/2);
	gui_printf(" %s! Are you sure? [Y/N] ",m);
	gui_buffer[0]=0; strpad(gui_buffer,i);
	gui_locate((80-i)/2,GUI_HEIGHT/2-1);
	gui_output(gui_buffer,i);
	gui_locate((80-i)/2,GUI_HEIGHT/2+1);
	gui_output(gui_buffer,i);
	do
	{
		push_pp(); gui_listen(); pop_pp();
		if (key_char==13||key_code=='Y'||key_code=='S'||key_code=='O'||key_code=='J')
			return 1;
	}
	while (key_code!=GUI_ESCAPE&&!GUI_ALTF4()&&key_code!='N');
	return 0;
}
int gui_prompt(char *s,int l,int x,int y) // enter text: ZERO is ABORT
{
	int i,j,z;
	i=j=strlen(s);
	strcpy(t,s); // must not use gui_buffer, gui_redraw() writes on it!
	strpad(t,l);
	do
	{
		if (i<0)
			i=0;
		else if (i>j)
			i=j;
		gui_locate(x,y);
		gui_colour(0xF0);
		gui_output(t,i);
		if (i<l)
		{
			gui_colour(0x0F);
			gui_output(&t[i],1);
			if (i<l-1)
			{
				gui_colour(0xF0);
				gui_output(&t[i+1],l-i-1);
			}
		}
		push_pp(); gui_listen(); pop_pp();
		if (key_char==13)
		{
			t[j]=0;
			strcpy(s,t);
			return 1;
		}
		if (key_code==GUI_HOME||key_code==GUI_UP)
			i=0;
		else if (key_code==GUI_END||key_code==GUI_DOWN)
			i=j;
		else if (key_code==GUI_LEFT)
			--i;
		else if (key_code==GUI_RIGHT)
			++i;
		else if ((key_char==8&&i>0)||(key_code==GUI_DELETE&&i<j))
		{
			if (key_char==8)
				--i;
			for (z=i;z<l-1;++z)
				t[z]=t[z+1];
			t[z]=' ';
			--j;
		}
		else if (key_char>=' '&&j<l)
		{
			for (z=l-1;z>i;--z)
				t[z]=t[z-1];
			t[i++]=key_char;
			++j;
		}
	}
	while (key_code!=GUI_ESCAPE&&!GUI_ALTF4());
	return 0;
}

#define GUI_MAXDIRLEN (2<<9)
#define GUI_MAXONELEN (1<<4)
WIN32_FIND_DATA pFindData;
unsigned char gui_filechar[GUI_MAXDIRLEN*GUI_MAXONELEN],gui_filepath[256],gui_filefind[256],*gui_filename[GUI_MAXDIRLEN];
int gui_filesize[GUI_MAXDIRLEN],gui_knownunits=0;
static unsigned char gui_browseup[]="..\\",gui_padded80[]=" %-78s ";
#define gui_fillem(i,s) (((i)<GUI_MAXDIRLEN-1)&&((unsigned char *)(s)<&gui_filechar[sizeof(gui_filechar)-(GUI_MAXONELEN)]))
void gui_sortem(unsigned char **t,int l)
{
	// simple insertion sort
	int i,j; char *r,*s;
	for (i=1;i<l;++i)
	{
		s=t[j=i];
		while (j)
		{
			if (strcasecmp(r=t[j-1],s)<0)
				break;
			t[j--]=r;
		}
		t[j]=s;
	}
}
int gui_browse(char *fullpath,int savefile,char *dotext,char *caption) // select file: ZERO is ABORT
{
	gui_locate(0,GUI_HEIGHT/4-1);
	gui_colour(0xF8);
	gui_printf(gui_padded80,caption);
	int i,j,l=strlen(dotext),activefile,filenames;
	char *s,*r;
	HANDLE hFind;
	GetFullPathName(fullpath[0]?fullpath:"|",MAXSTRLEN,gui_filepath,&r);
	strcpy(gui_filefind,r);
	*r=0;
	for (;;) // build file list: gui_filepath="C:\PATH\", gui_filefind="FILENAME.EXT"
	{
		r=gui_filechar;
		filenames=0;
		// list drives
		strcpy(s=pFindData.cFileName,"A:\\"); i=0;
		while (pFindData.cFileName[0]<='Z')
		{
			if ((gui_knownunits>>i)&1)
			{
				gui_filesize[filenames]=0;
				gui_filename[filenames++]=r;
				*r++=*s;
				*r++=':';
				*r++='\\';
				*r++=0;
			}
			++i; ++pFindData.cFileName[0];
		}
		activefile=filenames;
		// list directories
		strcat(gui_filepath,"*");
		if (gui_fillem(filenames,r)&&(hFind=FindFirstFile(gui_filepath,&pFindData)))
		{
			j=filenames;
			do
			{
				if ((pFindData.dwFileAttributes&0x10)&&!(pFindData.dwFileAttributes&0x06))
					if (strcmp(s=pFindData.cFileName,"."))
					{
						gui_filesize[filenames]=0;
						gui_filename[filenames++]=r;
						while (*r++=*s++)
							{}
						--r;
						*r++='\\';
						*r++=0;
					}
			}
			while (gui_fillem(filenames,r)&&FindNextFile(hFind,&pFindData));
			FindClose(hFind);
			gui_sortem(&gui_filename[j],filenames-j);
		}
		// list files
		if (!savefile&&gui_filefind[0]=='|')
			activefile=filenames;
		strcat(gui_filepath,dotext);
		if (gui_fillem(filenames,r)&&(hFind=FindFirstFile(gui_filepath,&pFindData)))
		{
			j=filenames;
			do
			{
				if (!(pFindData.dwFileAttributes&0x16))
					if ((i=strlen(s=pFindData.cFileName)-l)>0&&!strcasecmp(&s[i],dotext))
					{
						gui_filesize[filenames]=pFindData.nFileSizeLow;
						gui_filename[filenames++]=r;
						while (*r++=*s++)
							{}
					}
			}
			while (gui_fillem(filenames,r)&&FindNextFile(hFind,&pFindData));
			FindClose(hFind);
			gui_sortem(&gui_filename[j],filenames-j);
		}
		if (savefile)
		{
			if (gui_filefind[0]=='|')
				activefile=filenames;
			s="<new file>";
			gui_filename[filenames++]=r;
			while (*r++=*s++)
				{}
		}
		if (gui_filefind[0]!='|')
		{
			i=activefile;
			while (i<filenames&&strcasecmp(gui_filename[i],gui_filefind))
				++i;
			if (i<filenames)
				activefile=i;
		}
		*r=0; // end of list
		int update=1;
		for (;;) // show list and handle keys
		{
			keepwithin(&activefile,0,filenames-1);
			if (update)
			{
				for (i=GUI_HEIGHT/4;i<=GUI_HEIGHT*3/4;++i)
				{
					gui_locate(0,i);
					j=i-((GUI_HEIGHT-1)/2)+activefile;
					if (j==activefile)
						gui_colour(0x0F);
					else
						gui_colour(0xF0);
					if (j>=0&&j<filenames)
						gui_printf(strlen(gui_filename[j])>68?" %.66s..%10i ":" %-68s%10i ",gui_filename[j],gui_filesize[j]);
					else
						gui_printf(gui_padded80,"");
				}
				gui_locate(0,i);
				gui_colour(0xF8);
				gui_printf(gui_padded80,gui_filepath);
			}
			update=0;
			push_pp(); gui_listen(); pop_pp();
			if (key_code==GUI_ESCAPE||GUI_ALTF4())
				return 0;
			if (key_char==13)
				break;
			switch(key_code)
			{
				case GUI_UP:
					--activefile;
					update=1;
					break;
				case GUI_DOWN:
					++activefile;
					update=1;
					break;
				case GUI_LEFT:
				case GUI_PRIOR:
					activefile-=GUI_HEIGHT/2;
					update=1;
					break;
				case GUI_RIGHT:
				case GUI_NEXT:
					activefile+=GUI_HEIGHT/2;
					update=1;
					break;
				case GUI_HOME:
					activefile=0;
					update=1;
					break;
				case GUI_END:
					activefile=filenames-1;
					update=1;
					break;
				default:
					key_char=toupper(key_char);
					for (i=activefile+1;i<filenames;++i)
						if (toupper(gui_filename[i][0])==key_char)
							break;
					if (i>=filenames)
						for (i=0;i<activefile;++i)
							if (toupper(gui_filename[i][0])==key_char)
								break;
					if (i!=activefile)
					{
						activefile=i;
						update=1;
					}
					break;
			}
		}
		// shall we rebuild the list?
		if (gui_filename[activefile][0]=='<')
		{
			gui_filename[activefile][0]=0;
			if (!gui_prompt(gui_filename[activefile],78-12,1,(GUI_HEIGHT-1)/2))
				return 0;
			if (!(r=strrchr(gui_filename[activefile],'.'))||(r<strrchr(gui_filename[activefile],'\\')))
				strcat(gui_filename[activefile],dotext);
			GetFullPathName(gui_filepath,256,(char *)pFindData.cFileName,&r);
			strcpy(r,gui_filename[activefile]);
			break;
		}
		else if (gui_filename[activefile][1]==':')
		{
			strcpy(gui_filepath,gui_filename[activefile]);
			gui_filefind[0]=0;
		}
		else if (strchr(gui_filename[activefile],'\\'))
		{
			if (strcmp(gui_filename[activefile],gui_browseup))
				strcpy(gui_filefind,gui_browseup);
			else
			{
				r=s=strrchr(gui_filepath,'\\');
				while (*--s!='\\')
					{}
				*++r=0;
				strcpy(gui_filefind,++s);
				//*r='\\';
			}
			strcpy((char *)pFindData.cFileName,gui_filepath);
			r=strrchr((char *)pFindData.cFileName,'\\');
			strcpy(++r,gui_filename[activefile]);
			GetFullPathName((char *)pFindData.cFileName,256,gui_filepath,&r);
		}
		else
		{
			GetFullPathName(gui_filepath,256,(char *)pFindData.cFileName,&r);
			strcpy(r,gui_filename[activefile]);
			break;
		}
	}
	strcpy(fullpath,(char *)pFindData.cFileName);
	return 1;
}

int extendscores(int i)
{
	int j=0;
	if (scrln[order[0][activeorder][0]]<i)
		j=4,scrln[order[0][activeorder][0]]=i;
	if (scrln[order[1][activeorder][0]]<i)
		j=4,scrln[order[1][activeorder][0]]=i;
	if (scrln[order[2][activeorder][0]]<i)
		j=4,scrln[order[2][activeorder][0]]=i;
	return j;
}
#define extendinstrs(i) if (instrs<i) instrs=i;
#define extendorders(i) if (orders<i) orders=i;

int transposeall(void) // apply current transposition to whole song
{
	if (transpose)
	{
		undo_set();
		int m;
		for (m=0;m<orders;++m)
		{
			order[0][m][1]+=transpose;
			order[1][m][1]+=transpose;
			order[2][m][1]+=transpose;
		}
		transpose=0;
		return 4|8;
	}
	return 0;
}

int comparescores(unsigned char order1[128][2],unsigned char order2[128][2],int l)
{
	int i=0,j,k,d=0,n=0;
	while (l--)
	{
		j=order1[i][0];
		k=order2[i][0];
		if (!j||!k||j>(CHIPNSFX_SIZE-2)||k>(CHIPNSFX_SIZE-2))
		{
			if (j!=k)
				return 128; // give up if it isn't note VS note!
		}
		else
		{
			if (order1[i][1]!=order2[i][1])
				return 128; // different instruments, give up!
			if (j>(CHIPNSFX_SIZE-4)&&k>(CHIPNSFX_SIZE-4))
				{}
			else
			{
				if (j>(CHIPNSFX_SIZE-4)||k>(CHIPNSFX_SIZE-4))
					return 128;
				if (j+d!=k)
				{
					if (n)
						return 128; // give up if delta changes twice!
					else
						d=k-j; // set delta
				}
				++n;
			}
		}
		++i;
	}
	return d; // delta
}

int dupe_instr_test(int q)
{
	int i,j;
	for (j=0;j<256;++j)
		for (i=0;i<scrln[j];++i)
			if (score[j][i][1]==q)
				return j;
	return j; // unused instrument
}
int dupe_instr_show(int q)
{
	if (q<256)
		q=gui_readmef("\nInstrument appears in pattern #%02X.\n",q);
	else
		q=gui_readme("\nUnused instrument.\n");
	return q;
}

int charshortcuts(char k)
{
	switch(k)
	{
		case '+':
			selectorder=++activeorder;
			selectorderitem=activeorderitem;
			return 4|1;
		case '-':
			selectorder=--activeorder;
			selectorderitem=activeorderitem;
			return 4|1;
		case '*':
			++activeoctave;
			return 8;
		case '/':
			--activeoctave;
			return 8;
		case '<':
			--activeinstr;
			return 2;
		case '>':
			++activeinstr;
			return 2;
	}
	return 0;
}
int validhex(char k)
{
	if ((k>='0')&&(k<='9'))
		return k-'0';
	if ((k>='A')&&(k<='F'))
		return k-'A'+10;
	return -1;
}

char *guierrprintf(char *s)
{
	sprintf(t,"\nError: %s!\n",s);
	return t;
}

#endif

/*--------*/

int stderrprintf(char *s)
{
	fprintf(stderr,"error: %s!\n",s);
	return 1;
}

int optchr=1; // internal index within string
char *optarg; int optind=1,optopt; // opterr not supported!
INLINE int getopt(int argc, char **argv, char *optstring)
{
	for (;;)
	{
		if (optchr<1) // skip?
			++optind,optchr=1;
		if (optind<1||optind>=argc)
			return -1; // out of range or end of line!
		if ((optarg=argv[optind])[0]!='-'||!optarg[1])
			return optchr=0; // non-standard (*!*) cases: "-" or an argument!
		if (optopt=optarg[optchr++])
			break;
		optchr=0; // skip
	}
	--argc;
	char *s=optstring,c;
	while (*s==':')
		++s; // ignore ':', like opterr
	while (c=*s++)
		if (c==optopt)
		{
			if (*s==':')
			{
				if (!(c=optarg[optchr])&&optind>=argc)
					return ':'; // missing argument!
				optarg=c?&argv[optind][optchr]:argv[++optind];
				optchr=0; // skip
			}
			return optopt; // OK!
		}
	return '?'; // unrecognized character!
}

static char gui_chp_ext[]=".chp",gui_load_song[]="Load song",txt_read_error[]="cannot read song",txt_wave_error[]="cannot save WAVE",txt_ym3b_error[]="cannot save YM3b",txt_save_error[]="cannot save song";

int main(int argc,char *argv[])
{
	int i,j,k,l,m,n,o,p,x,y,z=0,flag_p=0,flag_w=0,flag_y=0;
	char *si=NULL,*so=NULL;
	strcpy(title0,"song_");
	filepath[0]=0;
	while ((i=getopt(argc,argv,":vql:LsbB:qpwWmrRc:FQk:PyYaTN:Zn:UtX"))!=-1)
		switch (i)
		{
			case 'v':
				flag_v=1, flag_q=0;
				break;
			case 'q':
				flag_q=1, flag_v=0;
				break;
			case 'l':
				strcpy(title0,optarg);
				break;
			case 'L':
				flag_ll=1;
				break;
			case 's':
				z=1;
				break;
			case 'b':
				flag_b|=7*9; // 1+2+4
				break;
			case 'B':
				j=0;
				while (k=optarg[j++])
				{
					switch (k-'0')
					{
						case 1:
							flag_b|=1*9;
							break;
						case 2:
							flag_b|=2*9;
							break;
						case 3:
							flag_b|=4*9;
							break;
						case 4:
							flag_b|=1;
							break;
						case 5:
							flag_b|=2;
							break;
						case 6:
							flag_b|=4;
							break;
						case 7:
							if (k=(optarg[j]=='-'))
								++j;
							flag_bb[0]=k?2:1;
							flag_bb[1]=k?1:2;
							flag_bb[2]=0;
							break;
						case 8:
							if (k=(optarg[j]=='-'))
								++j;
							flag_bb[0]=k?2:0;
							flag_bb[1]=k?0:2;
							flag_bb[2]=1;
							break;
						case 9:
							if (k=(optarg[j]=='-'))
								++j;
							flag_bb[0]=k?1:0;
							flag_bb[1]=k?0:1;
							flag_bb[2]=2;
						case 0:
							break;
						default:
							optind=-1;
							break;
					}
				}
				break;
			case 'p':
				flag_p=1;
				break;
			case 'w':
				flag_w=1;
				break;
			case 'W':
				flag_w=2;
				break;
			case 'm':
				flag_m=1;
				break;
			case 'r':
				flag_r|=2;
				break;
			case 'R':
				flag_r|=1;
				break;
			case 'c':
				flag_c=atoi(optarg);
				break;
			case 'F':
				SND_QUALITY=SND_HIGHEST;
				break;
			case 'Q':
				++SND_DEPTH;
				break;
			case 'k':
				if (optarg[0]=='1')
					strcpy(pianos,pianoz[0]);
				else if (optarg[0]=='2')
					strcpy(pianos,pianoz[1]);
				else if (strlen(optarg)==strlen(pianos))
					strcpy(pianos,optarg);
				else
					optind=-1;
				break;
			case 'P':
				flag_pp=1;
				break;
			case 'y':
				flag_y=1;
				break;
			case 'Y':
				flag_y=2;
				break;
			case 'a':
				flag_a=1;
				break;
			case 'T':
				flag_t=2;
				break;
			case 'N':
				flag_nn=strtol(optarg,NULL,10);
				break;
			case 'Z':
				++flag_zz;
				break;
			case 'n':
				flag_n=atoi(optarg);
				break;
			case 'U':
				flag_uu=1;
				break;
			case 't':
				flag_t=1;
				break;
			case 'X':
				++flag_xx;
				break;
			case 0:
				if (!strcmp(optarg,"--help"))
					optind=-1;
				else if (!si)
					si=optarg;
				else if (!so)
					so=optarg;
				else
					optind=-1;
				break;
			default:
				optind=-1;
				break;
		}
	if (optind!=argc)
	{
		printf(

			MY_FILE " " MY_DATE " (C) " MY_SELF "\n\n"
			"usage: " MY_FILE " [option..] song [target]]\n"
			"\t-v\tverbose output\n"
			"\t-q\tquiet, no output\n"
			"\t-l STR\tset label prefix\n"
			"\t-L\tembed header\n"
			"\t-c N\tmaximum target width\n"
			"\t-b\tgenerate short calls only\n"
			"\t-B N\tfine-tune short calls for channel[s] N\n"
			"\t-r\talternative loop search\n"
			"\t-R\tdon't look for loops at all\n"
			"\t-t\tsimplify timings\n"
			"\t-T\tdon't store tempo at all\n"
			"\t-a\tappend to target\n"
			"\t-F\t48000 Hz playback\n"
			//"\t-Z\trectangular waves\n"
			//"\t-Q\tenable oversampling\n"
			"\t-w/-W\texport song/loop as WAVE\n"
			"\t-y/-Y\texport song/loop as YM3b\n"
			"\t-n N\texport N loops\n"
			"\t-N N\ttranspose N semitones\n"
			#ifndef DEBUG
			"\t-k1\tset German keymap\n"
			"\t-k2\tset French keymap\n"
			"\t-k STR\tset custom keymap\n"
			"\t-m\tmute, no sound\n"
			"\t-p\tplay song\n"
			//"\t-U\tdistinct Undo and Redo\n"
			"\t-P\tcursor follows playback\n"
			//"\t-X\tshow hexadecimal bars\n"
			#endif
			"\n" GPL_3_INFO "\n"
			);
		return 1;
	}
	if (flag_n<1)
		flag_n=1;
	flag_c-=10; // margin for ",$00FX,$XX"
	SND_QUANTUM=(SND_QUALITY+50)/100;
	keepwithin(&SND_DEPTH,1,5);
	SND_DEPTHBITS=(1<<SND_DEPTH)-1;
	newsong();
	if (si)
		if (!loadsong(si))
			return stderrprintf(txt_read_error);
	// commandline actions
	if (flag_w)
		return savewave(so,flag_w>1?loopto:0)?0:stderrprintf(txt_wave_error);
	else if (flag_y)
		return saveym3b(so,flag_y>1?loopto:0)?0:stderrprintf(txt_ym3b_error);
	else if (so||z)
		return (z?savesong(so):savedata(so))?0:stderrprintf(txt_save_error);
#ifndef DEBUG
	// interactive tracker
	if (!gui_init(80,25))
		return stderrprintf("cannot set terminal");
	gui_titles(filepath);
	snd_open();
	undo_set();
	flag_v=dirty=can_undo=can_redo=0;
	int update=flag_n=-1;
	if (flag_p&&hWave)
		snd_play(0);
	{
		char scratch[]="A:\\"; long int dummy,i=0;
		while (scratch[0]<='Z') // build cache at boot time, it's faster and avoids later conflicts
		{
			if (GetDriveType(scratch)>1&&GetDiskFreeSpace(scratch,&dummy,&dummy,&dummy,&dummy))
				gui_knownunits|=1<<i;
			++i; ++scratch[0];
		}
	}
	for (;;)
	{
		// boundaries
		keepwithin(&activeparameter,0,3-1);
		keepwithin(&activescoreitem,0,3-1);
		keepwithin(&activeinstritem,0,28-1);
		keepwithin(&activeorderitem,0,3-1);
		keepwithin(&activescoreitem,0,3-1);
		keepwithin(&activeinstritem,0,28-1);
		keepwithin(&activeorderitem,0,3-1);
		keepwithin(&activescore,0,128-1);
		keepwithin(&activeinstr,0,256-1);
		keepwithin(&activeorder,0,256-1);
		keepwithin(&activeoctave,0,8);
		keepwithin(&hertz,25,100);
		keepwithin(&divisor,1,99);
		keepwithin(&transpose,-99,+99);
		keepwithin(&selectscore,0,128-1);
		keepwithin(&selectscoreitem,0,3-1);
		keepwithin(&selectorderitem,0,3-1);
		gui_redraw(update&~8);
		update=0;
		gui_listen();
		int writeable=!(flag_pp&&snd_playing);
		if (key_code==GUI_ESCAPE||GUI_ALTF4())
		{
			update=-1;
			if (gui_really(dirty?"Quit without saving":"Quit"))
				break;
		}
		else switch(key_code)
		{
			case GUI_F1: // HELP
				if (key_flag&GUI_CONTROL)
				{
					update=16;
					snd_bool[0]=snd_bool[1]=snd_bool[2]=1;
				}
				else update=gui_readme(
					"\n-=- " MY_NAME " (C) " MY_SELF " -=-\n"
					"\n"
					"F2\tgo to pattern panel\t+ ^PU\tnext order\n"
					"F3\tgo to instrument panel\t- ^PD\tpast order\n"
					"F4\tgo to order list panel\t* ^Rt\tnext octave\n"
					"F5\tplay song\t\t/ ^Lt\tpast octave\n"
					"F6\tplay pattern\t\t> ^Dn\tnext instrument\n"
					"F7\tplay from cursor\t< ^Up\tpast instrument\n"
					"F8\tstop playback\t\tENTER\tget note instrument\n"
					"F9\tgo to parameter panel\tSPACE\tset note instrument\n"
					"F11\tedit song title\t\tF12\tedit description\n"
					"^F1..F4\ttoggle channel on or off (Shift: solo channel)\n"
					"^Q/A\traise/lower by a semitone (Shift: by an octave)\n"
					"^E/D\tlook for repeated elements (Shift: extend/erase)\n"
					"^T/G\tshift selected notes, instrs. or patterns up/down\n"
					"^W\tlevel transposition\t^U\tclean orders/patterns\n"
					"^K\tset last note/instr./pattern\t^P\ttoggle cursor\n"
					"^L\tset 'loop to' pattern\t^Z\tundo\t^Y\tredo\n"
					"^X\tcut\t^C\tcopy\t^V\tpaste\t^B\tappend\n"
					"^N\tnew song\t^O\tload song\tF1\thelp\n"
					"^R\treload song\t^S\tsave song\tESC\tquit\n"
					//" S D   G H J   2 3   5 6 7   9 0        1   Rest    .   Clear\n"
					//"Z X C V B N M Q W E R T Y U I O P       L   C-9 (white noise)\n"
					);
				break;
			case GUI_F2: // GO TO PATTERN PANEL
			case GUI_F3: // GO TO INSTRUMENT PANEL
			case GUI_F4: // GO TO ORDER LIST PANEL
				i=key_code-GUI_F2; // assuming GUI_F4=GUI_F3+1=GUI_F2+2
				if (key_flag&GUI_CONTROL)
				{
					update=16;
					if (key_flag&GUI_SHIFT)
					{
						snd_bool[0]=snd_bool[1]=snd_bool[2]=0;
						snd_bool[i]=1;
					}
					else
						snd_bool[i]=!snd_bool[i];
				}
				else
				{
					update=(1<<activepanel)|(1<<i)|16;
					activepanel=i;
				}
				break;
			case GUI_F9: // GO TO PARAMETER PANEL
				update=(1<<activepanel)|16;
				activepanel=3;
				break;
			case GUI_F5: // PLAY FULL SONG
			case GUI_F6: // PLAY PATTERN
			case GUI_F7: // PLAY FROM CURSOR
				if (hWave)
				{
					if (key_code==GUI_F7&&snd_playing&&flag_pp)
						snd_looping=0;
					else if (key_code==GUI_F6&&snd_playing&&flag_pp)
						snd_looping=1;
					else
					{
						snd_stop();
						snd_order=key_code==GUI_F5?((key_flag&GUI_CONTROL)&&loopto>0?loopto:0):activeorder;
						snd_score=key_code!=GUI_F7?0:activescore;
						snd_play(key_code==GUI_F6);
					}
				}
				break;
			case GUI_F8: // STOP PLAYBACK
				snd_stop();
				break;
			case GUI_F11: // EDIT SONG TITLE
				gui_prompt(title1,80-2,1,0);//39,1
				update=32;
				break;
			case GUI_F12: // EDIT DESCRIPTION
				gui_prompt(title2,80-2,1,0);//39,41
				update=32;
				break;
			case GUI_UP:
				if (key_flag&GUI_CONTROL)
				{
					update=2;
					--activeinstr;
				}
				else switch(activepanel)
				{
					case 0:
						if (writeable)
						{
							update=1;
							--activescore;
							if (!(key_flag&GUI_SHIFT))
							{
								selectscore=activescore;
								selectscoreitem=activescoreitem;
							}
						}
						break;
					case 1:
						update=2;
						--activeinstr;
						break;
					case 2:
						if (writeable)
						{
							update=4|1;
							--activeorder;
							if (!(key_flag&GUI_SHIFT))
							{
								selectorder=activeorder;
								selectorderitem=activeorderitem;
							}
						}
						break;
					case 3:
						switch(activeparameter)
						{
							case 0:
								hertz*=2;
								break;
							case 1:
								++divisor;
								break;
							case 2:
								++transpose;
								break;
						}
						break;
				}
				break;
			case GUI_DOWN:
				if (key_flag&GUI_CONTROL)
				{
					update=2;
					++activeinstr;
				}
				else switch(activepanel)
				{
					case 0:
						if (writeable)
						{
							update=1;
							++activescore;
							if (!(key_flag&GUI_SHIFT))
							{
								selectscore=activescore;
								selectscoreitem=activescoreitem;
							}
						}
						break;
					case 1:
						update=2;
						++activeinstr;
						break;
					case 2:
						if (writeable)
						{
							update=4|1;
							++activeorder;
							if (!(key_flag&GUI_SHIFT))
							{
								selectorder=activeorder;
								selectorderitem=activeorderitem;
							}
						}
						break;
					case 3:
						switch(activeparameter)
						{
							case 0:
								hertz/=2;
								break;
							case 1:
								--divisor;
								break;
							case 2:
								--transpose;
								break;
						}
						break;
				}
				break;
			case GUI_LEFT:
				if (key_flag&GUI_CONTROL)
				{
					--activeoctave;
				}
				else switch(activepanel)
				{
					case 0:
						update=1;
						--activescoreitem;
						if (!(key_flag&GUI_SHIFT))
						{
							selectscore=activescore;
							selectscoreitem=activescoreitem;
						}
						break;
					case 1:
						update=2;
						--activeinstritem;
						break;
					case 2:
						update=4;
						--activeorderitem;
						if (!(key_flag&GUI_SHIFT))
						{
							selectorder=activeorder;
							selectorderitem=activeorderitem;
						}
						break;
					case 3:
						--activeparameter;
						break;
				}
				break;
			case GUI_RIGHT:
				if (key_flag&GUI_CONTROL)
				{
					++activeoctave;
				}
				else switch(activepanel)
				{
					case 0:
						update=1;
						++activescoreitem;
						if (!(key_flag&GUI_SHIFT))
						{
							selectscore=activescore;
							selectscoreitem=activescoreitem;
						}
						break;
					case 1:
						update=2;
						++activeinstritem;
						break;
					case 2:
						update=4;
						++activeorderitem;
						if (!(key_flag&GUI_SHIFT))
						{
							selectorder=activeorder;
							selectorderitem=activeorderitem;
						}
						break;
					case 3:
						++activeparameter;
						break;
				}
				break;
			case GUI_PRIOR:
				if (key_flag&GUI_CONTROL)
				{
					if (writeable)
					{
						selectorder=--activeorder;
						selectorderitem=activeorderitem;
						update=4|1;
					}
					break;
				}
				update=1<<activepanel;
				switch(activepanel)
				{
					case 0:
						if (writeable)
						{
							activescore-=16;
							if (!(key_flag&GUI_SHIFT))
							{
								selectscore=activescore;
								selectscoreitem=activescoreitem;
							}
						}
						break;
					case 1:
						activeinstr-=16;
						break;
					case 2:
						if (writeable)
						{
							update|=1;
							activeorder-=16;
							if (!(key_flag&GUI_SHIFT))
							{
								selectorder=activeorder;
								selectorderitem=activeorderitem;
							}
						}
						break;
					case 3:
						switch(activeparameter)
						{
							case 0:
								hertz=100;
								break;
							case 1:
								divisor+=10;
								break;
							case 2:
								transpose+=12;
								break;
						}
						break;
				}
				break;
			case GUI_NEXT:
				if (key_flag&GUI_CONTROL)
				{
					if (writeable)
					{
						selectorder=++activeorder;
						selectorderitem=activeorderitem;
						update=4|1;
					}
					break;
				}
				update=1<<activepanel;
				switch(activepanel)
				{
					case 0:
						if (writeable)
						{
							activescore+=16;
							if (!(key_flag&GUI_SHIFT))
							{
								selectscore=activescore;
								selectscoreitem=activescoreitem;
							}
						}
						break;
					case 1:
						activeinstr+=16;
						break;
					case 2:
						if (writeable)
						{
							update|=1;
							activeorder+=16;
							if (!(key_flag&GUI_SHIFT))
							{
								selectorder=activeorder;
								selectorderitem=activeorderitem;
							}
						}
						break;
					case 3:
						switch(activeparameter)
						{
							case 0:
								hertz=25;
								break;
							case 1:
								divisor-=10;
								break;
							case 2:
								transpose-=12;
								break;
						}
						break;
				}
				break;
			case GUI_HOME:
				update=1<<activepanel;
				switch(activepanel)
				{
					case 0:
						if (writeable)
						{
							activescore=0;
							if (!(key_flag&GUI_SHIFT))
							{
								selectscore=activescore;
								selectscoreitem=activescoreitem;
							}
						}
						break;
					case 1:
						activeinstr=1;
						break;
					case 2:
						if (writeable)
						{
							update|=1;
							activeorder=(key_flag&GUI_ALT)?loopto:0;
							if (!(key_flag&GUI_SHIFT))
							{
								selectorder=activeorder;
								selectorderitem=activeorderitem;
							}
						}
						break;
					case 3:
						activeparameter=0;
						break;
				}
				break;
			case GUI_END:
				update=1<<activepanel;
				switch(activepanel)
				{
					case 0:
						if (writeable)
						{
							activescore=scrln[order[0][activeorder][0]]-1;
							if (!(key_flag&GUI_SHIFT))
							{
								selectscore=activescore;
								selectscoreitem=activescoreitem;
							}
						}
						break;
					case 1:
						activeinstr=instrs-1;
						break;
					case 2:
						if (writeable)
						{
							update|=1;
							activeorder=orders-1;
							if (!(key_flag&GUI_SHIFT))
							{
								selectorder=activeorder;
								selectorderitem=activeorderitem;
							}
						}
						break;
					case 3:
						activeparameter=3-1;
						break;
				}
				break;
			#define INTMINMAX_SCORE (INTMINMAX(m,n,activescore,selectscore),INTMINMAX(o,p,activescoreitem,selectscoreitem))
			#define INTMINMAX_ORDER (INTMINMAX(m,n,activeorder,selectorder),INTMINMAX(o,p,activeorderitem,selectorderitem))
			case GUI_INSERT:
			case GUI_DELETE:
				if (writeable)
				{
					k=(key_code==GUI_DELETE)?-1:+1;
					switch(activepanel)
					{
						case 0:
							update=1;
							undo_set();
							INTMINMAX_SCORE;
							for (j=o;j<=p;++j)
							{
								l=order[j][activeorder][0];
								if (k>0)
								{
									memmove(score[l][n+1],score[l][m],(128-1-n)*sizeof(score[l][m]));
									memset(score[l][m],0,(n-m+1)*sizeof(score[l][m]));
								}
								else
								{
									memmove(score[l][m],score[l][n+1],(128-1-n)*sizeof(score[l][m]));
									memset(score[l][128-1-n+m],0,(n-m+1)*sizeof(score[l][m]));
								}
							}
							break;
						case 1:
							if (k<0&&activeinstr>0&&activeinstritem<16)
								if (writeable)
								{
									update=2;
									undo_set();
									memmove(&instr[activeinstr][activeinstritem+8],
										&instr[activeinstr][activeinstritem+9],16-1-activeinstritem);
									instr[activeinstr][8+16-1]=' ';
								}
							break;
						case 2:
							if ((k>0&&orders<256)||(k<0&&orders>=activeorder))
							{
								update=4|1;
								undo_set();
								INTMINMAX_ORDER;
								for (j=o;j<=p;++j)
									if (k>0)
									{
										memmove(order[j][n+1],order[j][m],(256-1-n)*sizeof(order[l][m]));
										//memset(order[j][m],*?*,...);
									}
									else
									{
										memmove(order[j][m],order[j][n+1],(256-1-n)*sizeof(order[l][m]));
										//memset(order[j][256-1-n+m],*?*,...);
									}
								if (p-o==2)
								{
									orders+=k*(1+n-m);
									if (m<loopto)
										loopto+=k*(1+n-m);
								}
							}
							break;
					}
				}
				break;
			default:
				switch(key_char)
				{
					case  1: // C-a: LOWER BY SEMITONE/OCTAVE
					case 17: // C-q: RAISE BY SEMITONE/OCTAVE
						if (writeable)
						{
							k=(key_char==1)?-1:+1;
							if (key_flag&GUI_SHIFT)
								k*=12;
							update=activepanel==2?4:1;
							if (activepanel<3)
								undo_set();
							switch (activepanel)
							{
								case 0:
									INTMINMAX_SCORE;
									for (j=o;j<=p;++j)
										for (i=m,l=order[j][activeorder][0];i<=n;++i)
											score[l][i][0]=addnote(score[l][i][0],k);
									break;
								case 1:
									for (i=0;i<256;++i)
										for (j=0;j<128;++j)
											if (score[i][j][1]==activeinstr)
												score[i][j][0]=addnote(score[i][j][0],k);
									break;
								case 2:
									INTMINMAX_ORDER;
									for (j=o;j<=p;++j)
										for (i=m;i<=n;++i)
											order[j][i][1]=intmax(intmin(k+(signed char)order[j][i][1],+99),-99);
									break;
								case 3:
									transpose+=k;
									break;
							}
						}
						break;
					case 23: // C-w: "FLATTEN" SCORE TRANSPOSITION
						switch(activepanel)
						{
							case 3:
								update=transposeall();
								break;
							case 0:
							case 2:
								if (activepanel==0)
									m=n=activeorder,INTMINMAX(o,p,activescoreitem,selectscoreitem);
								else
									INTMINMAX_ORDER;
								for (j=o;j<=p;++j)
									for (i=m;i<=n;++i)
									{
										k=order[j][i][0];
										if (l=(signed char)order[j][i][1])
										{
											if (!update)
												undo_set();
											for (z=0;z<128;++z)
												score[k][z][0]=addnote(score[k][z][0],l);
											for (z=0;z<orders;++z)
											{
												if (order[0][z][0]==k)
													order[0][z][1]-=l;
												if (order[1][z][0]==k)
													order[1][z][1]-=l;
												if (order[2][z][0]==k)
													order[2][z][1]-=l;
											}
											update=4|1;
										}
									}
								break;
						}
						break;
					case  8: // C-h, BACKSPACE, SHIFT+TAB: RIGHT-TO-LEFT
						update=(1<<activepanel)|16;
						if (key_flag&GUI_CONTROL)
							update|=1<<(activepanel=(activepanel-1+4)%4);
						else switch(activepanel)
						{
							case 0:
								if (key_flag&GUI_SHIFT)
									--activescoreitem;
								else
									if (writeable)
										--activescore;
								selectscore=activescore;
								selectscoreitem=activescoreitem;
								break;
							case 1:
								if (key_flag&GUI_SHIFT)
									activeinstritem=activeinstritem<16+1?0:16;
								else
								{
									if (activeinstr>0&&activeinstritem>0)
										if (--activeinstritem<16)//-1
											if (writeable)
											{
												undo_set();
												memmove(&instr[activeinstr][activeinstritem+8],
													&instr[activeinstr][activeinstritem+9],16-1-activeinstritem);
												instr[activeinstr][8+16-1]=' ';
											}
								}
								break;
							case 2:
								update|=1;
								if (key_flag&GUI_SHIFT)
									--activeorderitem;
								else
									if (writeable)
										--activeorder;
								selectorder=activeorder;
								selectorderitem=activeorderitem;
								break;
							case 3:
								--activeparameter;
								break;
						}
						break;
					case  9: // C-i, TAB: LEFT-TO-RIGHT
						update=(1<<activepanel)|16;
						if (key_flag&GUI_CONTROL)
							update|=1<<(activepanel=(activepanel+1)%4);
						else switch(activepanel)
						{
							case 0:
								selectscoreitem=++activescoreitem;
								selectscore=activescore;
								break;
							case 1:
								activeinstritem=activeinstritem<16+1?16+1:16+11;
								break;
							case 2:
								selectorderitem=++activeorderitem;
								selectorder=activeorder;
								break;
							case 3:
								++activeparameter;
								break;
						}
						break;
					case 11: // C-k: SET LAST SCORE/INSTR/ORDER
						update=1<<activepanel;
						switch(activepanel)
						{
							case 0:
								update|=4; // fix "00+00!01+00!02+00"
								scrln[order[0][activeorder][0]]=
									scrln[order[1][activeorder][0]]=
									scrln[order[2][activeorder][0]]=
									activescore+1;
								break;
							case 1:
								instrs=activeinstr+1;
								break;
							case 2:
								orders=activeorder+1;
								break;
						}
						break;
					case 12: // C-l: SET "LOOPTO"
						update=4;
						loopto=loopto==activeorder?-1:activeorder;
						break;
					case 10: // C-j [!]
					case 13: // C-m, RETURN: MISCELLANEOUS, ETC.
						switch(activepanel)
						{
							case 0:
								if (key_flag&GUI_CONTROL)
								{
									if (key_flag&GUI_SHIFT) // selected notes = 1st NOTE of active INSTR
									{
										if (writeable)
										{
											undo_set();
											INTMINMAX(m,n,activescore,selectscore);
											memcpy(t,score[order[activescoreitem][activeorder][0]][m],(l=(n-m+1))*sizeof(score[0][0]));
											for (j=0;j<256;++j)
												for (i=0;i<(128+1-l);++i)
												{
													z=score[j][i][0]-t[0];
													for (k=0;k<l;++k)
													{
														if (score[j][i+k][1]!=t[k*sizeof(score[0][0])+sizeof(score[0][0][0])])
															break;
														x=score[j][i+k][0]; y=t[k*sizeof(score[0][0])];
														if (!x||x>(CHIPNSFX_SIZE-4)||!y||y>(CHIPNSFX_SIZE-4))
														{
															if (x!=y)
																break;
														}
														else if ((x-y)!=z)
															break;
													}
													if (k>=l)
													{
														k=addnote(score[j][i][0],12*(activeoctave-4));
														memset(score[j][i],0,l*sizeof(score[0][0]));
														score[j][i][0]=k;
														score[j][i][1]=(!k||k>CHIPNSFX_SIZE-2)?0:activeinstr;
													}
												}
											update=1;
										}
									}
									else // DETECT ARPEGGIO
									{
										if (activescore+2<scrln[i=order[activescoreitem][activeorder][0]])
										{
											j=score[i][activescore+0][0];
											k=score[i][activescore+1][0]-j;
											l=score[i][activescore+2][0]-j;
											if (k>0&&k<=24&&l>=0&&l<=24)
												update=gui_readmef("\nArpeggio 00-%02i-%02i.\n",k,l);
										}
									}
								}
								else
								{
									update=2;
									activeinstr=score[order[activescoreitem][activeorder][0]][activescore][1];
								}
								break;
							case 1:
								if (activeinstritem>16)
								{
									snd_testn=CHIPNSFX_SIZE-3; // SFX: C-9
									snd_testd=0;
									snd_testi=activeinstr;
								}
								break;
						}
						break;
					case 14: // C-n: NEW SONG
						update=-1;
						if (!dirty||gui_really("New song"))
						{
							snd_stop();
							newsong();
							undo_set();
							filepath[0]=dirty=can_undo=can_redo=activescore=selectscore=activeorder=selectorder=0;
							activeinstr=1;
							gui_titles(filepath);
						}
						break;
					case 15: // C-o: LOAD SONG...
						update=-1;
						if (!dirty||gui_really(gui_load_song))
						{
							if (gui_browse(filepath,0,gui_chp_ext,gui_load_song))
							{
								snd_stop();
								if (!loadsong(filepath))
									gui_readme(guierrprintf(txt_read_error));
								else
								{
									undo_set();
									dirty=can_undo=can_redo=activescore=selectscore=activeorder=selectorder=0;
									activeinstr=1;
									gui_titles(filepath);
								}
							}
						}
						break;
					case 18: // C-r: RELOAD SONG
						update=-1;
						if (filepath[0]&&(!dirty||gui_really("Reload song")))
						{
							snd_stop();
							if (!loadsong(filepath))
								gui_readme(guierrprintf(txt_read_error));
							else
								undo_set(),dirty=can_undo=can_redo=0;
						}
						break;
					case 19: // C-s: SAVE SONG
						update=-1;
						if (gui_browse(filepath,1,gui_chp_ext,"Save song"))
						{
							if (!savesong(filepath))
								gui_readme(guierrprintf(txt_save_error));
							else
								dirty=0,gui_titles(filepath);
						}
						break;
					case  2: // C-b: PASTE AND ADVANCE
					case 22: // C-v: PASTE
						if (writeable)
							if (clipboardtype==activepanel&&(clipboards*clipboarditems))
							{
								undo_set();
								switch(activepanel)
								{
									case 0:
										update=1;
										m=intmin(activescore,selectscore);
										o=intmin(activescoreitem,selectscoreitem);
										for (j=0;j<clipboarditems&&j+o<3;++j)
											for (i=0,l=order[j+o][activeorder][0];i<clipboards&&i+m<128;++i)
												MEMLOAD(score[l][i+m],clipboard[j][i]);
										update|=extendscores(i+m);
										if (key_char!=22) // ADVANCE?
										{
											selectscore=activescore=i+m;
											selectscoreitem=activescoreitem=o;
										}
										break;
									case 2:
										update=4|1;
										m=intmin(activeorder,selectorder);
										o=intmin(activeorderitem,selectorderitem);
										for (i=0;i<clipboards&&(l=i+m)<256;++i)
											for (j=0;j<clipboarditems&&j+o<3;++j)
												MEMLOAD(order[j+o][l],clipboard[j][i]);
										extendorders(i+m);
										if (key_char!=22) // ADVANCE?
										{
											selectorder=activeorder=i+m;
											selectorderitem=activeorderitem=o;
										}
										break;
								}
							}
						break;
					case  3: // C-c: COPY
					case 24: // C-x: CUT
						if (writeable)
						{
							clipboardtype=activepanel;
							switch(activepanel)
							{
								case 0:
									clipboards=(n=intmax(activescore,selectscore))-(m=intmin(activescore,selectscore))+1;
									clipboarditems=(p=intmax(activescoreitem,selectscoreitem))-(o=intmin(activescoreitem,selectscoreitem))+1;
									for (j=0;j<clipboarditems;++j)
										for (i=0,l=order[j+o][activeorder][0];i<clipboards;++i)
											MEMLOAD(clipboard[j][i],score[l][i+m]);
									if (key_char==24) // CUT?
									{
										update=1;
										undo_set();
										for (j=o;j<=p;++j)
											for (i=m,l=order[j][activeorder][0];i<=n;++i)
												score[l][i][0]=score[l][i][1]=0;
									}
									break;
								case 2:
									clipboards=(n=intmax(activeorder,selectorder))-(m=intmin(activeorder,selectorder))+1;
									clipboarditems=(p=intmax(activeorderitem,selectorderitem))-(o=intmin(activeorderitem,selectorderitem))+1;
									for (i=0;i<clipboards;++i)
										for (j=0,l=i+m;j<clipboarditems;++j)
											MEMLOAD(clipboard[j][i],order[j+o][l]);
									if (key_char==24) // CUT?
									{
										update=4;
										undo_set();
										for (i=m;i<=n;++i)
											for (j=o;j<=p;++j)
												order[j][i][0]=order[j][i][1]=0;
									}
									break;
							}
						}
						break;
					case 20: // C-t: SHIFT UP
					case  7: // C-g: SHIFT DOWN
						if (writeable)
						{
							k=(key_char==20)?-1:+1;
							switch(activepanel)
							{
								case 0:
									INTMINMAX_SCORE;
									if ((k<0&&m+k>=0)||(k>0&&n<scrln[order[activescoreitem][activeorder][0]]-k))
									{
										update=1;
										undo_set();
										if (key_flag&GUI_SHIFT)
											k=k<0?-m:scrln[order[activescoreitem][activeorder][0]]-n-1;
										int gap,src,tgt;
										if (k<0)
											gap=-k,src=m+k,tgt=n+1+k;
										else
											gap=k,src=n+1,tgt=m;
										for (j=o;j<=p;++j)
										{
											l=order[j][activeorder][0];
											memcpy(t,score[l][src],sizeof(score[l][m])*gap);
											memmove(score[l][m+k],score[l][m],sizeof(score[l][m])*(n-m+1));
											memcpy(score[l][tgt],t,sizeof(score[l][m])*gap);
										}
										selectscore+=k;
										activescore+=k;
									}
									break;
								case 1:
									if ((k<0&&activeinstr>0-k)||(k>0&&activeinstr>0&&activeinstr<instrs-k))
									{
										update=2+1;
										undo_set();
										if (key_flag&GUI_SHIFT)
										{
											if (k<0)
												k=1-activeinstr;
											else
												k=instrs-activeinstr-1;
										}
										int gap,src,tgt;
										if (k<0)
											gap=-k,src=activeinstr+k,tgt=activeinstr+1+k;
										else
											gap=k,src=activeinstr+1,tgt=activeinstr;
										o=intmin(activeinstr,activeinstr+k);
										p=intmax(activeinstr,activeinstr+k);
										memcpy(t,instr[activeinstr],sizeof(instr[0]));
										memmove(instr[tgt],instr[src],sizeof(instr[0])*gap);
										memcpy(instr[activeinstr+k],t,sizeof(instr[0]));
										for (i=0;i<256;++i)
											for (j=0;j<128;++j)
												if ((l=score[i][j][1])==activeinstr)
													score[i][j][1]+=k;
												else if (l>=o&&l<=p)
													score[i][j][1]-=k<0?-1:1;
										activeinstr+=k;
									}
									break;
								case 2:
									INTMINMAX_ORDER;
									if ((k<0&&m+k>=0)||(k>0&&n<orders-k))
									{
										update=4;
										undo_set();
										if (key_flag&GUI_SHIFT)
										{
											if (k<0)
												k=-m;
											else
												k=orders-n-1;
										}
										int gap,src,tgt;
										if (k<0)
											gap=-k,src=m+k,tgt=n+1+k;
										else
											gap=k,src=n+1,tgt=m;
										for (j=o;j<=p;++j)
										{
											memcpy(t,order[j][src],sizeof(order[j][m])*gap);
											memmove(order[j][m+k],order[j][m],sizeof(order[j][m])*(n-m+1));
											memcpy(order[j][tgt],t,sizeof(order[j][m])*gap);
										}
										selectorder+=k;
										activeorder+=k;
									}
									break;
							}
						}
						break;
					case 25: // C-y: REDO
						update=-1;
						undo_get(1);
						break;
					case 26: // C-z: UNDO
						update=-1;
						undo_get(0);
						break;
					case  4: // C-d: DETECT USELESS DATA
						switch(activepanel)
						{
							case 1:
								if (activeinstr>0&&activeinstr<instrs)
								{
									j=dupe_instr_test(activeinstr);
									if (key_flag&GUI_SHIFT) // remove unused instrument
									{
										if (j<256)
											{}
										else
										{
											strcln(&instr[activeinstr][8]);
											for (i=j=0;i<9;++i)
												j|=instr[activeinstr][i];
											if (j)
											{
												undo_set();
												MEMVOID(instr[activeinstr]);
												update=2;
											}
										}
									}
									else
										update=dupe_instr_show(j);
								}
								break;
							case 0:
							case 2:
								if (activeorder<orders)
								{
									// LOOK FOR IDENTICAL SCORES
									if (activepanel==0)
										m=n=activeorder,INTMINMAX(o,p,activescoreitem,selectscoreitem);
									else
										INTMINMAX_ORDER;
									for (j=o;j<=p;++j)
										for (i=m;i<=n;++i)
										{
											k=order[j][i][0];
											for (z=0;z<m;++z)
											{
												if (k!=(y=order[0][z][0])&&(l=scrln[y])==scrln[k]&&(x=comparescores(score[y],score[k],l))!=128)
													break;
												if (k!=(y=order[1][z][0])&&(l=scrln[y])==scrln[k]&&(x=comparescores(score[y],score[k],l))!=128)
													break;
												if (k!=(y=order[2][z][0])&&(l=scrln[y])==scrln[k]&&(x=comparescores(score[y],score[k],l))!=128)
													break;
											}
											if (z<m)
											{
												if (key_flag&GUI_SHIFT) // merge identical pattern
												{
													if (!update)
														undo_set();
													order[j][i][0]=y;
													order[j][i][1]+=x;
													update=4|1;
												}
												else
												{
													update=gui_readmef("\nPattern #%02X is equal to #%02X%+03i.\n",k,y,(signed char)(order[j][i][1]+x));
													i=j=256; // force break
												}
											}
										}
									if (!update)
										update=gui_readme("\nNo previous duplicates.\n");
								}
								break;
						}
						break;
					case  5: // C-e: EXTEND CURRENT DATA
						switch(activepanel)
						{
							case 1:
								if (activeinstr>0&&activeinstr<instrs)
								{
									j=dupe_instr_test(activeinstr);
									if (key_flag&GUI_SHIFT) // clone instrument
									{
										if (j<256)
										{
											if (instrs<255)
											{
												undo_set();
												MEMLOAD(instr[instrs],instr[activeinstr]);
												activeinstr=instrs++;
												update=2;
											}
											else
												update=gui_readme("\nToo many instruments!\n");
										}
									}
									else
										update=dupe_instr_show(j);
								}
								break;
							case 0:
							case 2:
								if (activeorder<orders)
								{
									// LOOK FOR SHARED REFERENCES
									if (activepanel==0)
										m=n=activeorder,INTMINMAX(o,p,activescoreitem,selectscoreitem);
									else
										INTMINMAX_ORDER;
									for (j=o;j<=p;++j)
										for (i=m;i<=n;++i)
										{
											k=order[j][i][0];
											if (key_flag&GUI_SHIFT) // FORCE!
												z=-1;
											else
												for (z=0;z<m;++z)
													if (k==order[0][z][0]||k==order[1][z][0]||k==order[2][z][0])
														break;
											if (z<m)
											{
												if (key_flag&GUI_SHIFT)
												{
													// LOOK FOR UNUSED PATTERNS
													MEMVOID(save_hits);
													for (y=0;y<256;++y)
														save_hits[order[0][y][0]]=save_hits[order[1][y][0]]=save_hits[order[2][y][0]]=1;
													for (y=0;y<256;++y)
														if (!save_hits[y])
															break;
													if (y<256)
													{
														// EXTEND!
														if (!update)
															undo_set();
														MEMLOAD(score[y],score[k]);
														scrln[y]=scrln[k];
														order[j][i][0]=y;
														update=4|1;
													}
													else
													{
														update=gui_readme("\nToo many patterns!\n");
														i=j=256; // force break
													}
												}
												else
												{
													update=gui_readmef("\nPattern first used in order #%02X.\n",z);
													i=j=256; // force break
												}
											}
										}
									if (!update)
										update=gui_readme("\nNo previous references.\n");
								}
								break;
						}
						break;
					case 21: // C-u: PURGE ALL REPEATED PATTERNS, SORT AND CLEAN ORDERS
						switch(activepanel)
						{
							case 3:
								update=transposeall();
								break;
							case 0:
								// how many times is each pattern used?
								MEMVOID(save_hits);
								for (j=0;j<3;++j)
									for (i=0;i<orders;++i)
										++save_hits[order[j][i][0]];
								// FIND AND MERGE DUPLICATES: warning, not always optimal!
								for (o=p=0,i=1;i<256;++i)
									if (save_hits[i])
										for (j=0;j<i;++j)
											if (save_hits[j])
												if ((l=scrln[i])&&l==scrln[j]&&(m=comparescores(score[i],score[j],l))!=128)
													for (++o,k=0;k<orders;++k)
														for (n=0;n<3;++n)
															if (order[n][k][0]==i)
															{
																if (key_flag&GUI_SHIFT)
																	if (!p)
																		undo_set();
																++p;
																if (key_flag&GUI_SHIFT)
																{
																	order[n][k][0]=j;
																	order[n][k][1]-=m;
																	if (!--save_hits[i])
																	{
																		MEMVOID(score[i]);
																		scrln[i]=0;
																	}
																}
															}
								if (key_flag&GUI_SHIFT)
									update=1|4;
								else
									update=gui_readmef("\n%i pattern duplicates.\n",p);
								break;
							case 2:
								// LOOK FOR CLASHES AND FIX THEM, IF ANY
								for (j=k=0;j<3;++j)
									for (i=0;i<orders;++i)
										if ((l=order[j][i][0])>k)
											i=j=256; // force break
										else if (l==k)
										{
											if ((signed char)order[j][i][1])
												i=j=256; // force break
											else
												++k;
										}
								if (key_flag&GUI_SHIFT)
								{
									if (j>3)
									{
										undo_set();
										for (j=k=0;j<3;++j)
											for (i=0;i<orders;++i)
											{
												if ((l=order[j][i][0])>k) // sort entries
												{
													memswp(score[l],score[k],sizeof(score[0]));
													m=scrln[l];
													scrln[l]=scrln[k];
													scrln[k]=m;
													for (p=0;p<3;++p)
														for (o=0;o<orders;++o)
															if ((m=order[p][o][0])==k)
																order[p][o][0]=l;
															else if (m==l)
																order[p][o][0]=k;
													l=k;
												}
												if (l==k)
												{
													if (n=((signed char)order[j][i][1])) // fix transpositions
													{
														for (m=0;m<128;++m)
															score[l][m][0]=addnote(score[l][m][0],n);
														for (m=0;m<orders;++m)
														{
															if (order[0][m][0]==l)
																order[0][m][1]-=n;
															if (order[1][m][0]==l)
																order[1][m][1]-=n;
															if (order[2][m][0]==l)
																order[2][m][1]-=n;
														}
													}
													++k;
												}
											}
										update=4|1;
									}
								}
								else
									update=gui_readmef("\nOrders are %s sorted.\n",(j>3)?"already":"not");
								break;
						}
						break;
					case  6: // C-f: TOGGLE FLAG_ZZ
						if (key_flag&GUI_SHIFT)
							--flag_zz;
						else
							++flag_zz;
						break;
					case 16: // C-p: TOGGLE FLAG_PP
						if (key_flag&GUI_SHIFT)
							flag_xx=!flag_xx;
						else
							flag_pp=!flag_pp;
						break;
					case  0: // <!>
					case 27: // C-[ (ESC)
					case 28: // C-\ (FS)
					case 29: // C-] (GS)
					case 30: // C-^ (RS)
					case 31: // C-_ (US)
						break;
					default: // text characters
						if (writeable)
						{
							k=toupper(key_char);
							switch(activepanel)
							{
								case 0:
									if (k==' ') // set instrument on notes
									{
										undo_set();
										INTMINMAX_SCORE;
										for (j=o;j<=p;++j)
											for (i=m,l=order[j][activeorder][0];i<=n;++i)
												if ((k=score[l][i][0])&&k<(CHIPNSFX_SIZE-2))
													score[l][i][1]=activeinstr;
										activescore=selectscore=i;
										activescoreitem=selectscoreitem=o;
										update=1;
									}
									else if ((n=strchr(pianos,tolower(key_char))-pianos)>=0) // type note
									{
										if (n<12+12+6)
											k=n+activeoctave*12+(n<=12), // NOTE
											k=k<=(CHIPNSFX_SIZE-4)?k:(CHIPNSFX_SIZE-3); // SFX
										else if (n==12+12+6+0)
											k=CHIPNSFX_SIZE-3; // SFX
										else if (n==12+12+6+2)
											k=CHIPNSFX_SIZE-2; // BRAKE
										else if (n==12+12+6+1)
											k=CHIPNSFX_SIZE-0; // REST
										else
											k=0; // ERASE
										undo_set();
										update=(k&&k<(CHIPNSFX_SIZE-2))?activeinstr:0;
										INTMINMAX_SCORE;
										for (j=o;j<=p;++j)
											for (i=m,l=order[j][activeorder][0];i<=n;++i)
											{
												score[l][i][0]=k;
												score[l][i][1]=update;
											}
										snd_testn=k;
										snd_testd=(signed char)order[activescoreitem][activeorder][1];
										snd_testi=update;
										activescore=selectscore=i;
										activescoreitem=selectscoreitem=o;
										update=1|extendscores(activescore);
									}
									else if (i=charshortcuts(k))
										update=i;
									break;
								case 1:
									if (activeinstritem>16)
									{
										if (i=charshortcuts(k))
											update=i;
										else if ((l=validhex(k))>=0) // type instrument digits
										{
											update=2;
											undo_set();
											j=activeinstritem-17;
											if (j>7)
												++j;
											i=instr[activeinstr][j/2];
											if (j%2)
												i=(i&0xF0)+l;
											else
												i=(i&0x0F)+16*l;
											instr[activeinstr][j/2]=i;
											++activeinstritem;
											extendinstrs(activeinstr+1);
											instr[activeinstr][4]&=1;
										}
										else if (k==' ') // play A-octave
										{
											snd_testn=1+(9+12*activeoctave);
											snd_testd=0;
											snd_testi=activeinstr;
										}
									}
									else if (activeinstritem<16) // type instrument name
									{
										update=2;
										undo_set();
										strpad(&instr[activeinstr][8],16);
										memmove(&instr[activeinstr][activeinstritem+9],
											&instr[activeinstr][activeinstritem+8],16-1-activeinstritem);
										instr[activeinstr][activeinstritem+++8]=key_char;
										extendinstrs(activeinstr+1);
									}
									break;
								case 2:
									if ((l=validhex(k))>=0) // type order digits
									{
										update=4|1;
										undo_set();
										order[activeorderitem][activeorder][0]=
											order[activeorderitem][activeorder][0]*16+l;
										extendorders(activeorder+1);
									}
									else if (k==' ') // write instrument on patterns
									{
										update=4|1;
										undo_set();
										INTMINMAX_ORDER;
										for (j=o;j<=p;++j)
											for (i=m;i<=n;++i)
												for (l=0,z=order[j][i][0];l<128;++l)
													if ((k=score[z][l][0])&&k<(CHIPNSFX_SIZE-2))
														score[z][l][1]=activeinstr;
										activeorder=selectorder=i;
										activeorderitem=selectorderitem=o;
									}
									else if (i=charshortcuts(k))
										update=i;
									break;
								case 3:
									if (i=charshortcuts(k))
										update=i;
									break;
							}
						}
				}
		}
	}
	gui_exit();
	snd_close();
	return 0;
#else
	return stderrprintf("nothing to do");
#endif
}
